"""
Tests for ``preset_cli.api.clients.superset``.
"""
# pylint: disable=too-many-lines, trailing-whitespace

import json
from io import BytesIO
from unittest import mock
from uuid import UUID
from zipfile import ZipFile, is_zipfile

import pytest
import yaml
from pytest_mock import MockerFixture
from requests_mock.mocker import Mocker
from yarl import URL

from preset_cli import __version__
from preset_cli.api.clients.superset import (
    RuleType,
    SupersetClient,
    convert_to_adhoc_column,
    convert_to_adhoc_metric,
    parse_html_array,
)
from preset_cli.api.operators import OneToMany
from preset_cli.auth.main import Auth
from preset_cli.exceptions import ErrorLevel, SupersetError


def test_run_query(requests_mock: Mocker) -> None:
    """
    Test the ``run_query`` method.
    """
    requests_mock.post(
        "https://superset.example.org/superset/sql_json/",
        json={
            "query_id": 2,
            "status": "success",
            "data": [{"value": 1}],
            "columns": [{"name": "value", "type": "INT", "is_date": False}],
            "selected_columns": [{"name": "value", "type": "INT", "is_date": False}],
            "expanded_columns": [],
            "query": {
                "changedOn": "2022-03-25T15:37:00.393660",
                "changed_on": "2022-03-25T15:37:00.393660",
                "dbId": 1,
                "db": "examples",
                "endDttm": 1648222620417.808,
                "errorMessage": None,
                "executedSql": "SELECT 1 AS value\nLIMIT 1001",
                "id": "IrwwY8Ky14",
                "queryId": 2,
                "limit": 1000,
                "limitingFactor": "NOT_LIMITED",
                "progress": 100,
                "rows": 1,
                "schema": "public",
                "ctas": False,
                "serverId": 2,
                "sql": "SELECT 1 AS value",
                "sqlEditorId": "1",
                "startDttm": 1648222620279.198000,
                "state": "success",
                "tab": "Untitled Query 1",
                "tempSchema": None,
                "tempTable": None,
                "userId": 1,
                "user": "admin admin",
                "resultsKey": None,
                "trackingUrl": None,
                "extra": {"cancel_query": 35121, "progress": None},
            },
        },
    )

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)

    results = client.run_query(database_id=1, sql="SELECT 1 AS value", limit=10)
    assert results.to_dict() == {"value": {0: 1}}


def test_run_query_error(requests_mock: Mocker) -> None:
    """
    Test the ``run_query`` method when an error occurs.
    """
    errors = [
        {
            "message": "Only SELECT statements are allowed against this database.",
            "error_type": "DML_NOT_ALLOWED_ERROR",
            "level": "error",
            "extra": {
                "issue_codes": [
                    {
                        "code": 1022,
                        "message": (
                            "Issue 1022 - Database does not allow data manipulation."
                        ),
                    },
                ],
            },
        },
    ]
    requests_mock.post(
        "https://superset.example.org/superset/sql_json/",
        json={"errors": errors},
        headers={"Content-Type": "application/json"},
        status_code=400,
    )

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)

    with pytest.raises(SupersetError) as excinfo:
        client.run_query(database_id=1, sql="SSELECT 1 AS value", limit=10)
    assert excinfo.value.errors == [
        {
            "message": "Only SELECT statements are allowed against this database.",
            "error_type": "DML_NOT_ALLOWED_ERROR",
            "level": ErrorLevel.ERROR,
            "extra": {
                "issue_codes": [
                    {
                        "code": 1022,
                        "message": "Issue 1022 - Database does not allow data manipulation.",
                    },
                ],
            },
        },
    ]


def test_convert_to_adhoc_metric(mocker: MockerFixture) -> None:
    """
    Test ``convert_to_adhoc_metric``.
    """
    mocker.patch("preset_cli.api.clients.superset.uuid4", return_value=1234)
    assert convert_to_adhoc_metric("COUNT(*)") == {
        "aggregate": None,
        "column": None,
        "expressionType": "SQL",
        "hasCustomLabel": False,
        "isNew": False,
        "label": "COUNT(*)",
        "optionName": "metric_1234",
        "sqlExpression": "COUNT(*)",
    }


def test_convert_to_adhoc_column() -> None:
    """
    Test ``convert_to_adhoc_column`..
    """
    assert convert_to_adhoc_column("UPPER(name)") == {
        "label": "UPPER(name)",
        "sqlExpression": "UPPER(name)",
    }


def test_get_data(requests_mock: Mocker) -> None:
    """
    Test the ``run_query`` method.
    """
    requests_mock.get(
        "https://superset.example.org/api/v1/dataset/27",
        json={
            "description_columns": {},
            "id": 27,
            "label_columns": {
                "cache_timeout": "Cache Timeout",
                "columns.changed_on": "Columns Changed On",
                "columns.column_name": "Columns Column Name",
                "columns.created_on": "Columns Created On",
                "columns.description": "Columns Description",
                "columns.expression": "Columns Expression",
                "columns.extra": "Columns Extra",
                "columns.filterable": "Columns Filterable",
                "columns.groupby": "Columns Groupby",
                "columns.id": "Columns Id",
                "columns.is_active": "Columns Is Active",
                "columns.is_dttm": "Columns Is Dttm",
                "columns.python_date_format": "Columns Python Date Format",
                "columns.type": "Columns Type",
                "columns.type_generic": "Columns Type Generic",
                "columns.uuid": "Columns Uuid",
                "columns.verbose_name": "Columns Verbose Name",
                "database.backend": "Database Backend",
                "database.database_name": "Database Database Name",
                "database.id": "Database Id",
                "datasource_type": "Datasource Type",
                "default_endpoint": "Default Endpoint",
                "description": "Description",
                "extra": "Extra",
                "fetch_values_predicate": "Fetch Values Predicate",
                "filter_select_enabled": "Filter Select Enabled",
                "id": "Id",
                "is_sqllab_view": "Is Sqllab View",
                "main_dttm_col": "Main Dttm Col",
                "metrics": "Metrics",
                "offset": "Offset",
                "owners.first_name": "Owners First Name",
                "owners.id": "Owners Id",
                "owners.last_name": "Owners Last Name",
                "owners.username": "Owners Username",
                "schema": "Schema",
                "sql": "Sql",
                "table_name": "Table Name",
                "template_params": "Template Params",
                "url": "Url",
            },
            "result": {
                "cache_timeout": None,
                "columns": [
                    {
                        "changed_on": "2022-03-27T13:21:33.957609",
                        "column_name": "ts",
                        "created_on": "2022-03-27T13:21:33.957602",
                        "description": None,
                        "expression": None,
                        "extra": None,
                        "filterable": True,
                        "groupby": True,
                        "id": 841,
                        "is_active": True,
                        "is_dttm": True,
                        "python_date_format": None,
                        "type": "TIMESTAMP WITHOUT TIME ZONE",
                        "type_generic": 2,
                        "uuid": "e607d7fd-90bf-4420-a35e-9dc7af555e0d",
                        "verbose_name": None,
                    },
                    {
                        "changed_on": "2022-03-27T13:21:33.958499",
                        "column_name": "name",
                        "created_on": "2022-03-27T13:21:33.958493",
                        "description": None,
                        "expression": None,
                        "extra": None,
                        "filterable": True,
                        "groupby": True,
                        "id": 842,
                        "is_active": True,
                        "is_dttm": False,
                        "python_date_format": None,
                        "type": "VARCHAR(255)",
                        "type_generic": 1,
                        "uuid": "76a523f0-1aad-4608-87a4-daf22172e1da",
                        "verbose_name": None,
                    },
                    {
                        "changed_on": "2022-03-27T13:21:33.975750",
                        "column_name": "text",
                        "created_on": "2022-03-27T13:21:33.975743",
                        "description": None,
                        "expression": None,
                        "extra": None,
                        "filterable": True,
                        "groupby": True,
                        "id": 843,
                        "is_active": True,
                        "is_dttm": False,
                        "python_date_format": None,
                        "type": "TEXT",
                        "type_generic": 1,
                        "uuid": "610b91b0-e8de-4703-bbe9-5b27fb6c6a4e",
                        "verbose_name": None,
                    },
                ],
                "database": {
                    "backend": "postgresql",
                    "database_name": "superset_examples_dev",
                    "id": 3,
                },
                "datasource_type": "table",
                "default_endpoint": None,
                "description": "",
                "extra": json.dumps(
                    {
                        "resource_type": "model",
                        "unique_id": "model.superset_examples.messages_channels",
                        "depends_on": "ref('messages_channels')",
                    },
                ),
                "fetch_values_predicate": None,
                "filter_select_enabled": False,
                "id": 27,
                "is_sqllab_view": False,
                "main_dttm_col": "ts",
                "metrics": [
                    {
                        "changed_on": "2022-03-27T13:21:34.298657",
                        "created_on": "2022-03-27T13:21:34.248023",
                        "d3format": None,
                        "description": "",
                        "expression": "count(*)",
                        "extra": None,
                        "id": 35,
                        "metric_name": "cnt",
                        "metric_type": "count",
                        "uuid": "c4b74ceb-a19c-494a-9b90-9c68ed5bc8cb",
                        "verbose_name": "count(*)",
                        "warning_text": None,
                    },
                ],
                "offset": 0,
                "owners": [],
                "schema": "public",
                "sql": None,
                "table_name": "messages_channels",
                "template_params": None,
                "url": "/tablemodelview/edit/27",
            },
            "show_columns": [
                "id",
                "database.database_name",
                "database.id",
                "table_name",
                "sql",
                "filter_select_enabled",
                "fetch_values_predicate",
                "schema",
                "description",
                "main_dttm_col",
                "offset",
                "default_endpoint",
                "cache_timeout",
                "is_sqllab_view",
                "template_params",
                "owners.id",
                "owners.username",
                "owners.first_name",
                "owners.last_name",
                "columns.changed_on",
                "columns.column_name",
                "columns.created_on",
                "columns.description",
                "columns.expression",
                "columns.filterable",
                "columns.groupby",
                "columns.id",
                "columns.is_active",
                "columns.extra",
                "columns.is_dttm",
                "columns.python_date_format",
                "columns.type",
                "columns.uuid",
                "columns.verbose_name",
                "metrics",
                "datasource_type",
                "url",
                "extra",
                "columns.type_generic",
                "database.backend",
            ],
            "show_title": "Show Sqla Table",
        },
    )
    requests_mock.post(
        "https://superset.example.org/api/v1/chart/data",
        json={
            "result": [
                {
                    "cache_key": "ddf314a4bb44dd7f4a2f5b880e3c2503",
                    "cached_dttm": None,
                    "cache_timeout": 300,
                    "applied_template_filters": [],
                    "annotation_data": {},
                    "error": None,
                    "is_cached": None,
                    "query": """SELECT name AS name,
       count(*) AS cnt
FROM public.messages_channels
GROUP BY name
LIMIT 10000;

""",
                    "status": "success",
                    "stacktrace": None,
                    "rowcount": 29,
                    "from_dttm": None,
                    "to_dttm": None,
                    "colnames": ["name", "cnt"],
                    "indexnames": [
                        0,
                        1,
                        2,
                        3,
                        4,
                        5,
                        6,
                        7,
                        8,
                        9,
                        10,
                        11,
                        12,
                        13,
                        14,
                        15,
                        16,
                        17,
                        18,
                        19,
                        20,
                        21,
                        22,
                        23,
                        24,
                        25,
                        26,
                        27,
                        28,
                    ],
                    "coltypes": [1, 0],
                    "data": [
                        {"name": "newsletter", "cnt": 8},
                        {"name": "support", "cnt": 114},
                        {"name": "dashboard-level-access", "cnt": 57},
                        {"name": "beginners", "cnt": 368},
                        {"name": "community-feedback", "cnt": 27},
                        {"name": "graduation", "cnt": 38},
                        {"name": "superset_stage_alerts", "cnt": 3},
                        {"name": "contributing", "cnt": 29},
                        {"name": "github-notifications", "cnt": 2975},
                        {"name": "helm-k8-deployment", "cnt": 72},
                        {"name": "globalnav_search", "cnt": 1},
                        {"name": "design", "cnt": 4},
                        {"name": "localization", "cnt": 7},
                        {"name": "commits", "cnt": 1},
                        {"name": "apache-releases", "cnt": 53},
                        {"name": "jobs", "cnt": 22},
                        {"name": "general", "cnt": 383},
                        {"name": "announcements", "cnt": 19},
                        {"name": "visualization_plugins", "cnt": 50},
                        {"name": "product_feedback", "cnt": 41},
                        {"name": "dashboard-filters", "cnt": 75},
                        {"name": "superset-champions", "cnt": 40},
                        {"name": "introductions", "cnt": 141},
                        {"name": "embedd-dashboards", "cnt": 10},
                        {"name": "cypress-tests", "cnt": 7},
                        {"name": "superset_prod_reports", "cnt": 3},
                        {"name": "feature-requests", "cnt": 22},
                        {"name": "dashboards", "cnt": 87},
                        {"name": "developers", "cnt": 27},
                    ],
                    "result_format": "json",
                    "applied_filters": [],
                    "rejected_filters": [],
                },
            ],
        },
    )

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)

    results = client.get_data(27, ["cnt"], ["name"])
    assert results.to_dict() == {
        "name": {
            0: "newsletter",
            1: "support",
            2: "dashboard-level-access",
            3: "beginners",
            4: "community-feedback",
            5: "graduation",
            6: "superset_stage_alerts",
            7: "contributing",
            8: "github-notifications",
            9: "helm-k8-deployment",
            10: "globalnav_search",
            11: "design",
            12: "localization",
            13: "commits",
            14: "apache-releases",
            15: "jobs",
            16: "general",
            17: "announcements",
            18: "visualization_plugins",
            19: "product_feedback",
            20: "dashboard-filters",
            21: "superset-champions",
            22: "introductions",
            23: "embedd-dashboards",
            24: "cypress-tests",
            25: "superset_prod_reports",
            26: "feature-requests",
            27: "dashboards",
            28: "developers",
        },
        "cnt": {
            0: 8,
            1: 114,
            2: 57,
            3: 368,
            4: 27,
            5: 38,
            6: 3,
            7: 29,
            8: 2975,
            9: 72,
            10: 1,
            11: 4,
            12: 7,
            13: 1,
            14: 53,
            15: 22,
            16: 383,
            17: 19,
            18: 50,
            19: 41,
            20: 75,
            21: 40,
            22: 141,
            23: 10,
            24: 7,
            25: 3,
            26: 22,
            27: 87,
            28: 27,
        },
    }


def test_get_data_parameters(mocker: MockerFixture) -> None:
    """
    Test different parameters passed to ``get_data``.
    """
    auth = mocker.MagicMock()
    session = auth.get_session()
    session.get().json.return_value = {
        "result": {
            "columns": [],
            "metrics": [],
        },
    }
    session.post().json.return_value = {
        "result": [
            {
                "data": [{"a": 1}],
            },
        ],
    }
    mocker.patch("preset_cli.api.clients.superset.uuid4", return_value=1234)

    client = SupersetClient("https://superset.example.org/", auth)
    client.get_data(
        27,
        ["cnt"],
        ["name"],
        time_column="ts",
        is_timeseries=True,
        granularity="P1M",
    )

    session.post.assert_has_calls(
        [
            mock.call(),
            mock.call(
                URL("https://superset.example.org/api/v1/chart/data"),
                json={
                    "datasource": {"id": 27, "type": "table"},
                    "force": False,
                    "queries": [
                        {
                            "annotation_layers": [],
                            "applied_time_extras": {},
                            "columns": [{"label": "name", "sqlExpression": "name"}],
                            "custom_form_data": {},
                            "custom_params": {},
                            "extras": {
                                "having": "",
                                "having_druid": [],
                                "time_grain_sqla": "P1M",
                                "where": "",
                            },
                            "filters": [],
                            "granularity": "ts",
                            "is_timeseries": True,
                            "metrics": [
                                {
                                    "aggregate": None,
                                    "column": None,
                                    "expressionType": "SQL",
                                    "hasCustomLabel": False,
                                    "isNew": False,
                                    "label": "cnt",
                                    "optionName": "metric_1234",
                                    "sqlExpression": "cnt",
                                },
                            ],
                            "order_desc": True,
                            "orderby": [],
                            "row_limit": 10000,
                            "time_range": "No filter",
                            "timeseries_limit": 0,
                            "url_params": {},
                        },
                    ],
                    "result_format": "json",
                    "result_type": "full",
                },
                headers={
                    "Accept": "application/json",
                    "Content-Type": "application/json",
                    "User-Agent": f"Apache Superset Client ({__version__})",
                    "Referer": "https://superset.example.org/",
                },
            ),
            mock.call().ok.__bool__(),
            mock.call().json(),
        ],
    )


def test_get_data_time_column_error(requests_mock: Mocker) -> None:
    """
    Test when the time column is ambiguous in ``get_data``.
    """
    requests_mock.get(
        "https://superset.example.org/api/v1/dataset/27",
        json={
            "result": {
                "columns": [
                    {"column_name": "event_time", "is_dttm": True},
                    {"column_name": "server_time", "is_dttm": True},
                ],
                "metrics": [],
            },
        },
    )
    requests_mock.post(
        "https://superset.example.org/api/v1/chart/data",
        json={
            "result": [
                {
                    "data": [{"a": 1}],
                },
            ],
        },
    )

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)
    with pytest.raises(Exception) as excinfo:
        client.get_data(27, ["cnt"], ["name"])
    assert str(excinfo.value) == (
        "Unable to determine time column, please pass `time_series` "
        "as one of: event_time, server_time"
    )


def test_get_data_error(requests_mock: Mocker) -> None:
    """
    Test ``get_data`` with a generic error.
    """
    requests_mock.get(
        "https://superset.example.org/api/v1/dataset/27",
        json={
            "result": {
                "columns": [],
                "metrics": [],
            },
        },
    )
    requests_mock.post(
        "https://superset.example.org/api/v1/chart/data",
        json={"errors": [{"message": "An error occurred"}]},
        headers={"Content-Type": "application/json"},
        status_code=500,
    )

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)
    with pytest.raises(SupersetError) as excinfo:
        client.get_data(27, ["cnt"], ["name"], time_column="ts")
    assert excinfo.value.errors == [{"message": "An error occurred"}]


def test_get_resource(requests_mock: Mocker) -> None:
    """
    Test the generic ``get_resource`` method.
    """
    # the payload schema is irrelevant, since it's passed through unmodified
    requests_mock.get(
        "https://superset.example.org/api/v1/database/1",
        json={"Hello": "world"},
    )

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)

    response = client.get_resource("database", 1)
    assert response == {"Hello": "world"}
    assert (
        requests_mock.last_request.headers["Referer"] == "https://superset.example.org/"
    )


def test_get_resources(requests_mock: Mocker) -> None:
    """
    Test the generic ``get_resources`` method.
    """
    # the payload schema is irrelevant, since it's passed through unmodified
    requests_mock.get(
        "https://superset.example.org/api/v1/database/?q="
        "(filters:!(),order_column:changed_on_delta_humanized,"
        "order_direction:desc,page:0,page_size:100)",
        json={"result": [1]},
    )
    requests_mock.get(
        "https://superset.example.org/api/v1/database/?q="
        "(filters:!(),order_column:changed_on_delta_humanized,"
        "order_direction:desc,page:1,page_size:100)",
        json={"result": [2]},
    )
    requests_mock.get(
        "https://superset.example.org/api/v1/database/?q="
        "(filters:!(),order_column:changed_on_delta_humanized,"
        "order_direction:desc,page:2,page_size:100)",
        json={"result": []},
    )

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)

    response = client.get_resources("database")
    assert response == [1, 2]
    assert (
        requests_mock.last_request.headers["Referer"] == "https://superset.example.org/"
    )


def test_get_resources_filtered_equal(requests_mock: Mocker) -> None:
    """
    Test the generic ``get_resources`` method with an equal filter.
    """
    # the payload schema is irrelevant, since it's passed through unmodified
    requests_mock.get(
        "https://superset.example.org/api/v1/database/?q="
        "(filters:!((col:database_name,opr:eq,value:my_db)),"
        "order_column:changed_on_delta_humanized,"
        "order_direction:desc,page:0,page_size:100)",
        json={"result": [{"Hello": "world"}]},
    )
    requests_mock.get(
        "https://superset.example.org/api/v1/database/?q="
        "(filters:!((col:database_name,opr:eq,value:my_db)),"
        "order_column:changed_on_delta_humanized,"
        "order_direction:desc,page:1,page_size:100)",
        json={"result": []},
    )

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)

    response = client.get_resources("database", database_name="my_db")
    assert response == [{"Hello": "world"}]
    assert (
        requests_mock.last_request.headers["Referer"] == "https://superset.example.org/"
    )


def test_get_resources_filtered_one_to_many(requests_mock: Mocker) -> None:
    """
    Test the generic ``get_resources`` method with a one-to-many filter.
    """
    # the payload schema is irrelevant, since it's passed through unmodified
    requests_mock.get(
        "https://superset.example.org/api/v1/database/?q="
        "(filters:!((col:database,opr:rel_o_m,value:1)),"
        "order_column:changed_on_delta_humanized,"
        "order_direction:desc,page:0,page_size:100)",
        json={"result": [{"Hello": "world"}]},
    )
    requests_mock.get(
        "https://superset.example.org/api/v1/database/?q="
        "(filters:!((col:database,opr:rel_o_m,value:1)),"
        "order_column:changed_on_delta_humanized,"
        "order_direction:desc,page:1,page_size:100)",
        json={"result": []},
    )

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)

    response = client.get_resources("database", database=OneToMany(1))
    assert response == [{"Hello": "world"}]
    assert (
        requests_mock.last_request.headers["Referer"] == "https://superset.example.org/"
    )


def test_create_resource(requests_mock: Mocker) -> None:
    """
    Test the generic ``create_resource`` method.
    """
    requests_mock.post(
        "https://superset.example.org/api/v1/database/",
        json={"Hello": "world"},
    )

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)

    response = client.create_resource(
        "database",
        database_name="my_db",
        sqlalchemy_uri="gsheets://",
    )
    assert response == {"Hello": "world"}
    assert (
        requests_mock.last_request.headers["Referer"] == "https://superset.example.org/"
    )
    assert requests_mock.last_request.json() == {
        "database_name": "my_db",
        "sqlalchemy_uri": "gsheets://",
    }


def test_update_resource(requests_mock: Mocker) -> None:
    """
    Test the generic ``update_resource`` method.
    """
    requests_mock.put(
        "https://superset.example.org/api/v1/database/1",
        json={"Hello": "world"},
    )

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)

    response = client.update_resource(
        resource_name="database",
        resource_id=1,
        database_name="my_other_db",
    )
    assert response == {"Hello": "world"}
    assert (
        requests_mock.last_request.headers["Referer"] == "https://superset.example.org/"
    )
    assert requests_mock.last_request.json() == {
        "database_name": "my_other_db",
    }


def test_update_resource_with_query_args(requests_mock: Mocker) -> None:
    """
    Test the generic ``update_resource`` method.
    """
    requests_mock.put(
        "https://superset.example.org/api/v1/database/1?override_columns=true",
        json={"Hello": "world"},
    )

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)

    response = client.update_resource(
        resource_name="database",
        resource_id=1,
        query_args={"override_columns": "true"},
        database_name="my_other_db",
    )
    assert response == {"Hello": "world"}
    assert (
        requests_mock.last_request.headers["Referer"] == "https://superset.example.org/"
    )
    assert requests_mock.last_request.json() == {
        "database_name": "my_other_db",
    }


def test_get_database(mocker: MockerFixture) -> None:
    """
    Test the ``get_database`` method.
    """
    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)
    get_resource = mocker.patch.object(client, "get_resource")

    client.get_database(1)
    get_resource.assert_called_with("database", 1)


def test_get_databases(mocker: MockerFixture) -> None:
    """
    Test the ``get_databases`` method.
    """
    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)
    get_resources = mocker.patch.object(client, "get_resources")

    client.get_databases()
    get_resources.assert_called_with("database")
    client.get_databases(database_name="my_db")
    get_resources.assert_called_with("database", database_name="my_db")


def test_create_database(mocker: MockerFixture) -> None:
    """
    Test the ``create_database`` method.
    """
    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)
    create_resource = mocker.patch.object(client, "create_resource")

    client.create_database(database_name="my_db", sqlalchemy_uri="gsheets://")
    create_resource.assert_called_with(
        "database",
        database_name="my_db",
        sqlalchemy_uri="gsheets://",
    )


def test_update_database(mocker: MockerFixture) -> None:
    """
    Test the ``update_database`` method.
    """
    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)
    update_resource = mocker.patch.object(client, "update_resource")

    client.update_database(1, database_name="my_other_db")
    update_resource.assert_called_with(
        "database",
        1,
        {"override_columns": "true"},
        database_name="my_other_db",
    )


def test_get_dataset(mocker: MockerFixture) -> None:
    """
    Test the ``get_dataset`` method.
    """
    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)
    get_resource = mocker.patch.object(client, "get_resource")

    client.get_dataset(1)
    get_resource.assert_called_with("dataset", 1)


def test_get_datasets(mocker: MockerFixture) -> None:
    """
    Test the ``get_datasets`` method.
    """
    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)
    get_resources = mocker.patch.object(client, "get_resources")

    client.get_datasets()
    get_resources.assert_called_with("dataset")
    client.get_datasets(dataset_name="my_db")
    get_resources.assert_called_with("dataset", dataset_name="my_db")


def test_create_dataset(mocker: MockerFixture) -> None:
    """
    Test the ``create_dataset`` method.
    """
    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)
    create_resource = mocker.patch.object(client, "create_resource")

    client.create_dataset(dataset_name="my_db", sqlalchemy_uri="gsheets://")
    create_resource.assert_called_with(
        "dataset",
        dataset_name="my_db",
        sqlalchemy_uri="gsheets://",
    )


def test_update_dataset(mocker: MockerFixture) -> None:
    """
    Test the ``update_dataset`` method.
    """
    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)
    update_resource = mocker.patch.object(client, "update_resource")

    client.update_dataset(1, dataset_name="my_other_db")
    update_resource.assert_called_with("dataset", 1, dataset_name="my_other_db")


def test_get_chart(mocker: MockerFixture) -> None:
    """
    Test the ``get_chart`` method.
    """
    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)
    get_resource = mocker.patch.object(client, "get_resource")

    client.get_chart(1)
    get_resource.assert_called_with("chart", 1)


def test_get_charts(mocker: MockerFixture) -> None:
    """
    Test the ``get_charts`` method.
    """
    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)
    get_resources = mocker.patch.object(client, "get_resources")

    client.get_charts()
    get_resources.assert_called_with("chart")
    client.get_charts(chart_name="my_db")
    get_resources.assert_called_with("chart", chart_name="my_db")


def test_get_dashboard(mocker: MockerFixture) -> None:
    """
    Test the ``get_dashboard`` method.
    """
    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)
    get_resource = mocker.patch.object(client, "get_resource")

    client.get_dashboard(1)
    get_resource.assert_called_with("dashboard", 1)


def test_get_dashboards(mocker: MockerFixture) -> None:
    """
    Test the ``get_dashboards`` method.
    """
    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)
    get_resources = mocker.patch.object(client, "get_resources")

    client.get_dashboards()
    get_resources.assert_called_with("dashboard")
    client.get_dashboards(dashboard_name="my_db")
    get_resources.assert_called_with("dashboard", dashboard_name="my_db")


def test_create_dashboard(mocker: MockerFixture) -> None:
    """
    Test the ``create_dashboard`` method.
    """
    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)
    create_resource = mocker.patch.object(client, "create_resource")

    client.create_dashboard(dashboard_name="my_db", sqlalchemy_uri="gsheets://")
    create_resource.assert_called_with(
        "dashboard",
        dashboard_name="my_db",
        sqlalchemy_uri="gsheets://",
    )


def test_update_dashboard(mocker: MockerFixture) -> None:
    """
    Test the ``update_dashboard`` method.
    """
    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)
    update_resource = mocker.patch.object(client, "update_resource")

    client.update_dashboard(1, dashboard_name="my_other_db")
    update_resource.assert_called_with("dashboard", 1, dashboard_name="my_other_db")


def test_export_zip(requests_mock: Mocker) -> None:
    """
    Test the ``export_zip`` method.
    """
    buf = BytesIO()
    with ZipFile(buf, "w") as bundle:
        with bundle.open("test.txt", "w") as output:
            output.write(b"Hello!")
    buf.seek(0)

    requests_mock.get(
        "https://superset.example.org/api/v1/database/export/?q=%21%281%2C2%2C3%29",
        content=buf.getvalue(),
    )

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)

    response = client.export_zip("database", [1, 2, 3])
    assert is_zipfile(response)
    with ZipFile(response) as bundle:
        assert bundle.namelist() == ["test.txt"]
        assert bundle.read("test.txt") == b"Hello!"

    assert (
        requests_mock.last_request.headers["Referer"] == "https://superset.example.org/"
    )


def test_export_zip_pagination(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Test the ``export_zip`` method with pagination.
    """
    page1 = BytesIO()
    with ZipFile(page1, "w") as bundle:
        with bundle.open("test1.txt", "w") as output:
            output.write(b"Hello!")
    page1.seek(0)

    page2 = BytesIO()
    with ZipFile(page2, "w") as bundle:
        with bundle.open("test2.txt", "w") as output:
            output.write(b"Bye!")
    page2.seek(0)

    requests_mock.get(
        "https://superset.example.org/api/v1/database/export/?q=%21%281%2C2%29",
        content=page1.getvalue(),
    )
    requests_mock.get(
        "https://superset.example.org/api/v1/database/export/?q=%21%283%29",
        content=page2.getvalue(),
    )

    mocker.patch("preset_cli.api.clients.superset.MAX_IDS_IN_EXPORT", new=2)

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)

    response = client.export_zip("database", [1, 2, 3])
    assert is_zipfile(response)
    with ZipFile(response) as bundle:
        assert bundle.namelist() == ["test1.txt", "test2.txt"]

    assert (
        requests_mock.last_request.headers["Referer"] == "https://superset.example.org/"
    )


def test_export_zip_error(requests_mock: Mocker) -> None:
    """
    Test the ``export_zip`` method when an error occurs.
    """
    requests_mock.get(
        "https://superset.example.org/api/v1/database/export/?q=%21%281%2C2%2C3%29",
        json={"errors": [{"message": "An error occurred"}]},
        headers={"Content-Type": "application/json"},
        status_code=500,
    )

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)

    with pytest.raises(SupersetError) as excinfo:
        client.export_zip("database", [1, 2, 3])
    assert excinfo.value.errors == [{"message": "An error occurred"}]
    assert (
        requests_mock.last_request.headers["Referer"] == "https://superset.example.org/"
    )


def test_import_zip(requests_mock: Mocker) -> None:
    """
    Test the ``import_zip`` method.
    """
    requests_mock.post(
        "https://superset.example.org/api/v1/database/import/",
        json={"message": "OK"},
    )

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)

    data = BytesIO("I'm a ZIP".encode("utf-8"))
    response = client.import_zip("database", data, overwrite=True)
    assert response is True
    assert (
        requests_mock.last_request.headers["Referer"] == "https://superset.example.org/"
    )
    assert requests_mock.last_request.headers["Accept"] == "application/json"

    boundary = (
        requests_mock.last_request.headers["Content-type"]
        .split(";")[1]
        .split("=")[1]
        .strip()
    )
    assert requests_mock.last_request.text == (
        f'--{boundary}\r\nContent-Disposition: form-data; name="overwrite"\r\n\r\n'
        f'true\r\n--{boundary}\r\nContent-Disposition: form-data; name="formData"; '
        f'filename="formData"\r\n\r\nI\'m a ZIP\r\n--{boundary}--\r\n'
    )


def test_import_zip_error(requests_mock: Mocker) -> None:
    """
    Test the ``import_zip`` method when an error occurs.
    """
    requests_mock.post(
        "https://superset.example.org/api/v1/database/import/",
        json={"errors": [{"message": "An error occurred"}]},
        headers={"Content-Type": "application/json"},
        status_code=500,
    )

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)

    data = BytesIO("I'm a ZIP".encode("utf-8"))
    with pytest.raises(SupersetError) as excinfo:
        client.import_zip("database", data, overwrite=True)
    assert excinfo.value.errors == [{"message": "An error occurred"}]
    assert (
        requests_mock.last_request.headers["Referer"] == "https://superset.example.org/"
    )


def test_export_users(requests_mock: Mocker) -> None:
    """
    Test ``export_users``.
    """
    requests_mock.get(
        "https://superset.example.org/users/list/?psize_UserDBModelView=100&page_UserDBModelView=0",
        text="""
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="utf-8">
  </head>
  <body>
    <table></table>
    <table>
      <tr>
        <th></th>
        <th>First Name</th>
        <th>Last Name</th>
        <th>User Name</th>
        <th>Email</th>
        <th>Is Active?</th>
        <th>Role</th>
      </tr>
      <tr>
        <td><a href="/users/edit/1">Edit</a></td>
        <td>Alice</td>
        <td>Doe</td>
        <td>adoe</td>
        <td>adoe@example.com</td>
        <td>True</td>
        <td>[Admin]</td>
      </tr>
      <tr>
        <td><a href="/users/edit/2">Edit</a></td>
        <td>Bob</td>
        <td>Doe</td>
        <td>bdoe</td>
        <td>bdoe@example.com</td>
        <td>True</td>
        <td>[Alpha]</td>
      </tr>
    </table>
  </body>
</html>
        """,
    )
    requests_mock.get(
        "https://superset.example.org/users/list/?psize_UserDBModelView=100&page_UserDBModelView=1",
        text="""
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="utf-8">
  </head>
  <body>
    <table></table>
    <table>
      <tr>
        <th></th>
        <th>First Name</th>
        <th>Last Name</th>
        <th>User Name</th>
        <th>Email</th>
        <th>Is Active?</th>
        <th>Role</th>
      </tr>
    </table>
  </body>
</html>
        """,
    )

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)
    assert list(client.export_users()) == [
        {
            "id": 1,
            "first_name": "Alice",
            "last_name": "Doe",
            "username": "adoe",
            "email": "adoe@example.com",
            "role": ["Admin"],
        },
        {
            "id": 2,
            "first_name": "Bob",
            "last_name": "Doe",
            "username": "bdoe",
            "email": "bdoe@example.com",
            "role": ["Alpha"],
        },
    ]


def test_export_rls(requests_mock: Mocker) -> None:
    """
    Test ``export_rls``.
    """
    requests_mock.get(
        (
            "https://superset.example.org/rowlevelsecurityfiltersmodelview/list/?"
            "psize_RowLevelSecurityFiltersModelView=100&"
            "page_RowLevelSecurityFiltersModelView=0"
        ),
        text="""
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="utf-8">
  </head>
  <body>
    <table></table>
    <table>
      <tr>
        <th></th>
        <th>Name</th>
        <th>Filter Type</th>
        <th>Tables</th>
        <th>Roles</th>
        <th>Clause</th>
        <th>Creator</th>
        <th>Modified</th>
      </tr>
      <tr>
        <td><input id="1" /></td>
        <td>My rule</td>
        <td>Regular</td>
        <td>[main.test_table]</td>
        <td>client_id = 9</td>
        <td>admin admin</td>
        <td>35 minutes ago</td>
      </tr>
    </table>
  </body>
</html>
        """,
    )
    requests_mock.get(
        (
            "https://superset.example.org/rowlevelsecurityfiltersmodelview/list/?"
            "psize_RowLevelSecurityFiltersModelView=100&"
            "page_RowLevelSecurityFiltersModelView=1"
        ),
        text="""
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="utf-8">
  </head>
  <body>
    <table></table>
    <table>
      <tr>
        <th></th>
        <th>Name</th>
        <th>Filter Type</th>
        <th>Tables</th>
        <th>Roles</th>
        <th>Clause</th>
        <th>Creator</th>
        <th>Modified</th>
      </tr>
    </table>
  </body>
</html>
        """,
    )
    requests_mock.get(
        "https://superset.example.org/rowlevelsecurityfiltersmodelview/show/1",
        text="""
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="utf-8">
  </head>
  <body>
    <table>
      <tr><th>Name</th><td>My Rule</td></tr>
      <tr><th>Description</th><td>This is a rule. There are many others like it, but this one is mine.</td></tr>
      <tr><th>Filter Type</th><td>Regular</td></tr>
      <tr><th>Tables</th><td>[main.test_table]</td></tr>
      <tr><th>Roles</th><td>[Gamma]</td></tr>
      <tr><th>Group Key</th><td>department</td></tr>
      <tr><th>Clause</th><td>client_id = 9</td></tr>
    </table>
  </body>
</html>
        """,
    )

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)
    assert list(client.export_rls()) == [
        {
            "name": "My Rule",
            "description": "This is a rule. There are many others like it, but this one is mine.",
            "filter_type": "Regular",
            "tables": ["main.test_table"],
            "roles": ["Gamma"],
            "group_key": "department",
            "clause": "client_id = 9",
        },
    ]


def test_export_ownership(mocker: MockerFixture) -> None:
    """
    Test ``export_ownership``.
    """
    mocker.patch.object(
        SupersetClient,
        "export_users",
        return_value=[
            {"id": 1, "email": "admin@example.com"},
            {"id": 2, "email": "adoe@example.com"},
        ],
    )
    mocker.patch.object(
        SupersetClient,
        "get_uuids",
        return_value={
            1: UUID("e0d20af0-cef9-4bdb-80b4-745827f441bf"),
        },
    )
    mocker.patch.object(
        SupersetClient,
        "get_resources",
        return_value=[
            {
                "slice_name": "My chart",
                "id": 1,
                "owners": [{"id": 1}, {"id": 2}],
            },
        ],
    )

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)
    assert list(client.export_ownership("chart")) == [
        {
            "name": "My chart",
            "owners": ["admin@example.com", "adoe@example.com"],
            "uuid": UUID("e0d20af0-cef9-4bdb-80b4-745827f441bf"),
        },
    ]


def test_get_uuids(requests_mock: Mocker) -> None:
    """
    Test the ``get_uuids`` method.
    """
    requests_mock.get(
        "https://superset.example.org/api/v1/chart/?q="
        "(filters:!(),order_column:changed_on_delta_humanized,"
        "order_direction:desc,page:0,page_size:100)",
        json={"result": [{"id": 1}, {"id": 2}, {"id": 3}]},
    )
    requests_mock.get(
        "https://superset.example.org/api/v1/chart/?q="
        "(filters:!(),order_column:changed_on_delta_humanized,"
        "order_direction:desc,page:1,page_size:100)",
        json={"result": []},
    )

    uuids = [
        "c9d100b8-4fa5-4b7a-8a71-9803b5343674",
        "2826c33b-7d13-4830-865e-d62630b20dee",
        "0ac7464e-14e7-4c54-ab22-7cbd4536fccc",
    ]
    for i in range(3):
        name = f"chart_export/chart/chart{i+1}.yaml"
        uuid = uuids[i]
        buf = BytesIO()
        with ZipFile(buf, "w") as bundle:
            with bundle.open("metadata.yaml", "w") as output:
                output.write(b"Hello!")
            with bundle.open(name, "w") as output:
                output.write(yaml.dump({"uuid": uuid}).encode())
        buf.seek(0)

        requests_mock.get(
            f"https://superset.example.org/api/v1/chart/export/?q=%21%28{i+1}%29",
            content=buf.getvalue(),
        )

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)

    assert client.get_uuids("chart") == {
        1: UUID("c9d100b8-4fa5-4b7a-8a71-9803b5343674"),
        2: UUID("2826c33b-7d13-4830-865e-d62630b20dee"),
        3: UUID("0ac7464e-14e7-4c54-ab22-7cbd4536fccc"),
    }


def test_parse_html_array() -> None:
    """
    Test ``parse_html_array``.
    """
    assert (
        parse_html_array(
            """
                    
                      [main.test_table]
                    
                    """,
        )
        == ["main.test_table"]
    )
    assert (
        parse_html_array(
            """
                
                    
                        
                            main.sales
                        
                            public.FCC 2018 Survey
""",
        )
        == ["main.sales", "public.FCC 2018 Survey"]
    )


def test_import_rls(requests_mock: Mocker) -> None:
    """
    Test the ``import_rls`` method.
    """
    requests_mock.get(
        "https://superset.example.org/api/v1/dataset/?q="
        "(filters:!((col:schema,opr:eq,value:main),"
        "(col:table_name,opr:eq,value:test_table)),"
        "order_column:changed_on_delta_humanized,"
        "order_direction:desc,page:0,page_size:100)",
        json={"result": [{"id": 1}]},
    )
    requests_mock.get(
        "https://superset.example.org/api/v1/dataset/?q="
        "(filters:!((col:schema,opr:eq,value:main),"
        "(col:table_name,opr:eq,value:test_table)),"
        "order_column:changed_on_delta_humanized,"
        "order_direction:desc,page:1,page_size:100)",
        json={"result": []},
    )
    requests_mock.get(
        "https://superset.example.org/roles/list/?_flt_0_name=Gamma",
        text="""
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="utf-8">
  </head>
  <body>
    <table></table>
    <table>
      <tr>
        <td></td>
        <td>Name</td>
      </tr>
      <tr>
        <td><input id="1" /></td>
        <td>Gamma</td>
      </tr>
    </table>
  </body>
</html>
        """,
    )
    requests_mock.post(
        "https://superset.example.org/rowlevelsecurityfiltersmodelview/add",
    )

    rls: RuleType = {
        "clause": "client_id = 9",
        "description": "Rule description",
        "filter_type": "Regular",
        "group_key": "department",
        "name": "Rule name",
        "roles": ["Gamma"],
        "tables": ["main.test_table"],
    }

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)
    client.import_rls(rls)

    assert requests_mock.last_request.text == (
        "name=Rule+name&"
        "description=Rule+description&"
        "filter_type=Regular&"
        "tables=1&"
        "roles=1&"
        "group_key=department&"
        "clause=client_id+%3D+9"
    )


def test_import_rls_no_schema(requests_mock: Mocker) -> None:
    """
    Test the ``import_rls`` method when the table has no schema.
    """
    requests_mock.get(
        "https://superset.example.org/api/v1/dataset/?q="
        "(filters:!((col:table_name,opr:eq,value:test_table)),"
        "order_column:changed_on_delta_humanized,"
        "order_direction:desc,page:0,page_size:100)",
        json={"result": [{"id": 1}]},
    )
    requests_mock.get(
        "https://superset.example.org/api/v1/dataset/?q="
        "(filters:!((col:table_name,opr:eq,value:test_table)),"
        "order_column:changed_on_delta_humanized,"
        "order_direction:desc,page:1,page_size:100)",
        json={"result": []},
    )
    requests_mock.get(
        "https://superset.example.org/roles/list/?_flt_0_name=Gamma",
        text="""
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="utf-8">
  </head>
  <body>
    <table></table>
    <table>
      <tr>
        <td></td>
        <td>Name</td>
      </tr>
      <tr>
        <td><input id="1" /></td>
        <td>Gamma</td>
      </tr>
    </table>
  </body>
</html>
        """,
    )
    requests_mock.post(
        "https://superset.example.org/rowlevelsecurityfiltersmodelview/add",
    )

    rls: RuleType = {
        "clause": "client_id = 9",
        "description": "Rule description",
        "filter_type": "Regular",
        "group_key": "department",
        "name": "Rule name",
        "roles": ["Gamma"],
        "tables": ["test_table"],
    }

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)
    client.import_rls(rls)

    assert requests_mock.last_request.text == (
        "name=Rule+name&"
        "description=Rule+description&"
        "filter_type=Regular&"
        "tables=1&"
        "roles=1&"
        "group_key=department&"
        "clause=client_id+%3D+9"
    )


def test_import_rls_no_table(requests_mock: Mocker) -> None:
    """
    Test the ``import_rls`` method when no tables are found.
    """
    requests_mock.get(
        "https://superset.example.org/api/v1/dataset/?q="
        "(filters:!((col:schema,opr:eq,value:main),"
        "(col:table_name,opr:eq,value:test_table)),"
        "order_column:changed_on_delta_humanized,"
        "order_direction:desc,page:0,page_size:100)",
        json={"result": []},
    )

    rls: RuleType = {
        "clause": "client_id = 9",
        "description": "Rule description",
        "filter_type": "Regular",
        "group_key": "department",
        "name": "Rule name",
        "roles": ["Gamma"],
        "tables": ["main.test_table"],
    }

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)

    with pytest.raises(Exception) as excinfo:
        client.import_rls(rls)
    assert str(excinfo.value) == "Cannot find table: main.test_table"


def test_import_rls_multiple_tables(requests_mock: Mocker) -> None:
    """
    Test the ``import_rls`` method when multiple tables are found.
    """
    requests_mock.get(
        "https://superset.example.org/api/v1/dataset/?q="
        "(filters:!((col:schema,opr:eq,value:main),"
        "(col:table_name,opr:eq,value:test_table)),"
        "order_column:changed_on_delta_humanized,"
        "order_direction:desc,page:0,page_size:100)",
        json={"result": [{"id": 1}, {"id": 2}]},
    )
    requests_mock.get(
        "https://superset.example.org/api/v1/dataset/?q="
        "(filters:!((col:schema,opr:eq,value:main),"
        "(col:table_name,opr:eq,value:test_table)),"
        "order_column:changed_on_delta_humanized,"
        "order_direction:desc,page:1,page_size:100)",
        json={"result": []},
    )

    rls: RuleType = {
        "clause": "client_id = 9",
        "description": "Rule description",
        "filter_type": "Regular",
        "group_key": "department",
        "name": "Rule name",
        "roles": ["Gamma"],
        "tables": ["main.test_table"],
    }

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)

    with pytest.raises(Exception) as excinfo:
        client.import_rls(rls)
    assert str(excinfo.value) == "More than one table found: main.test_table"


def test_import_rls_no_role(requests_mock: Mocker) -> None:
    """
    Test the ``import_rls`` method when no role is found.
    """
    requests_mock.get(
        "https://superset.example.org/api/v1/dataset/?q="
        "(filters:!((col:schema,opr:eq,value:main),"
        "(col:table_name,opr:eq,value:test_table)),"
        "order_column:changed_on_delta_humanized,"
        "order_direction:desc,page:0,page_size:100)",
        json={"result": [{"id": 1}]},
    )
    requests_mock.get(
        "https://superset.example.org/api/v1/dataset/?q="
        "(filters:!((col:schema,opr:eq,value:main),"
        "(col:table_name,opr:eq,value:test_table)),"
        "order_column:changed_on_delta_humanized,"
        "order_direction:desc,page:1,page_size:100)",
        json={"result": []},
    )
    requests_mock.get(
        "https://superset.example.org/roles/list/?_flt_0_name=Gamma",
        text="""
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="utf-8">
  </head>
  <body>
    <table></table>
    <table>
      <tr>
        <td></td>
        <td>Name</td>
      </tr>
    </table>
  </body>
</html>
        """,
    )

    rls: RuleType = {
        "clause": "client_id = 9",
        "description": "Rule description",
        "filter_type": "Regular",
        "group_key": "department",
        "name": "Rule name",
        "roles": ["Gamma"],
        "tables": ["main.test_table"],
    }

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)

    with pytest.raises(Exception) as excinfo:
        client.import_rls(rls)
    assert str(excinfo.value) == "Cannot find role: Gamma"


def test_import_rls_multiple_roles(requests_mock: Mocker) -> None:
    """
    Test the ``import_rls`` method when multiple roles are found.

    Should not happen.
    """
    requests_mock.get(
        "https://superset.example.org/api/v1/dataset/?q="
        "(filters:!((col:schema,opr:eq,value:main),"
        "(col:table_name,opr:eq,value:test_table)),"
        "order_column:changed_on_delta_humanized,"
        "order_direction:desc,page:0,page_size:100)",
        json={"result": [{"id": 1}]},
    )
    requests_mock.get(
        "https://superset.example.org/api/v1/dataset/?q="
        "(filters:!((col:schema,opr:eq,value:main),"
        "(col:table_name,opr:eq,value:test_table)),"
        "order_column:changed_on_delta_humanized,"
        "order_direction:desc,page:1,page_size:100)",
        json={"result": []},
    )
    requests_mock.get(
        "https://superset.example.org/roles/list/?_flt_0_name=Gamma",
        text="""
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="utf-8">
  </head>
  <body>
    <table></table>
    <table>
      <tr>
        <td></td>
        <td>Name</td>
      </tr>
      <tr>
        <td><input id="1" /></td>
        <td>Gamma</td>
      </tr>
      <tr>
        <td><input id="2" /></td>
        <td>Gamma</td>
      </tr>
    </table>
  </body>
</html>
        """,
    )

    rls: RuleType = {
        "clause": "client_id = 9",
        "description": "Rule description",
        "filter_type": "Regular",
        "group_key": "department",
        "name": "Rule name",
        "roles": ["Gamma"],
        "tables": ["main.test_table"],
    }

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)

    with pytest.raises(Exception) as excinfo:
        client.import_rls(rls)
    assert str(excinfo.value) == "More than one role found: Gamma"


def test_import_rls_anchor_role_id(requests_mock: Mocker) -> None:
    """
    Test the ``import_rls`` method when the role ID is in an anchor tag.
    """
    requests_mock.get(
        "https://superset.example.org/api/v1/dataset/?q="
        "(filters:!((col:schema,opr:eq,value:main),"
        "(col:table_name,opr:eq,value:test_table)),"
        "order_column:changed_on_delta_humanized,"
        "order_direction:desc,page:0,page_size:100)",
        json={"result": [{"id": 1}]},
    )
    requests_mock.get(
        "https://superset.example.org/api/v1/dataset/?q="
        "(filters:!((col:schema,opr:eq,value:main),"
        "(col:table_name,opr:eq,value:test_table)),"
        "order_column:changed_on_delta_humanized,"
        "order_direction:desc,page:1,page_size:100)",
        json={"result": []},
    )
    requests_mock.get(
        "https://superset.example.org/roles/list/?_flt_0_name=Gamma",
        text="""
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="utf-8">
  </head>
  <body>
    <table></table>
    <table>
      <tr>
        <td></td>
        <td>Name</td>
      </tr>
      <tr>
        <td><a href="/roles/edit/1">Edit</a></td>
        <td>Gamma</td>
      </tr>
    </table>
  </body>
</html>
        """,
    )
    requests_mock.post(
        "https://superset.example.org/rowlevelsecurityfiltersmodelview/add",
    )

    rls: RuleType = {
        "clause": "client_id = 9",
        "description": "Rule description",
        "filter_type": "Regular",
        "group_key": "department",
        "name": "Rule name",
        "roles": ["Gamma"],
        "tables": ["main.test_table"],
    }

    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)
    client.import_rls(rls)

    assert requests_mock.last_request.text == (
        "name=Rule+name&"
        "description=Rule+description&"
        "filter_type=Regular&"
        "tables=1&"
        "roles=1&"
        "group_key=department&"
        "clause=client_id+%3D+9"
    )


def test_import_ownership(mocker: MockerFixture, requests_mock: Mocker) -> None:
    """
    Test the ``import_ownership`` method.
    """
    requests_mock.put("https://superset.example.org/api/v1/dataset/1", json={})
    mocker.patch.object(
        SupersetClient,
        "export_users",
        return_value=[
            {"id": 1, "email": "admin@example.com"},
            {"id": 2, "email": "adoe@example.com"},
        ],
    )
    mocker.patch.object(
        SupersetClient,
        "get_uuids",
        return_value={
            1: UUID("e0d20af0-cef9-4bdb-80b4-745827f441bf"),
        },
    )
    auth = Auth()
    client = SupersetClient("https://superset.example.org/", auth)
    client.import_ownership(
        "dataset",
        [
            {
                "name": "test_table",
                "owners": ["admin@example.com", "adoe@example.com"],
                "uuid": "e0d20af0-cef9-4bdb-80b4-745827f441bf",
            },
        ],
    )

    assert requests_mock.last_request.json() == {"owners": [1, 2]}



"""
Tests for ``preset_cli.cli.superset.sync.dbt.datasets``.
"""
# pylint: disable=invalid-name

import json
from typing import List
from unittest import mock

import pytest
from pytest_mock import MockerFixture

from preset_cli.api.clients.dbt import MetricSchema, ModelSchema
from preset_cli.cli.superset.sync.dbt.datasets import sync_datasets

metric_schema = MetricSchema()
metrics: List[MetricSchema] = [
    metric_schema.load(
        {
            "depends_on": ["model.superset_examples.messages_channels"],
            "description": "",
            "filters": [],
            "meta": {},
            "name": "cnt",
            "label": "",
            "sql": "*",
            "type": "count",
            "unique_id": "metric.superset_examples.cnt",
        },
    ),
]

model_schema = ModelSchema()
models: List[ModelSchema] = [
    model_schema.load(
        {
            "database": "examples_dev",
            "schema": "public",
            "description": "",
            "meta": {},
            "name": "messages_channels",
            "unique_id": "model.superset_examples.messages_channels",
        },
    ),
]


def test_sync_datasets_new(mocker: MockerFixture) -> None:
    """
    Test ``sync_datasets`` when no datasets exist yet.
    """
    client = mocker.MagicMock()
    client.get_datasets.return_value = []
    client.create_dataset.side_effect = [{"id": 1}, {"id": 2}, {"id": 3}]

    sync_datasets(
        client=client,
        models=models,
        metrics=metrics,
        database={"id": 1},
        disallow_edits=False,
        external_url_prefix="",
    )
    client.create_dataset.assert_has_calls(
        [
            mock.call(database=1, schema="public", table_name="messages_channels"),
        ],
    )
    client.update_dataset.assert_has_calls(
        [
            mock.call(
                1,
                description="",
                extra=json.dumps(
                    {
                        "unique_id": "model.superset_examples.messages_channels",
                        "depends_on": "ref('messages_channels')",
                        "certification": {"details": "This table is produced by dbt"},
                    },
                ),
                is_managed_externally=False,
                metrics=[],
            ),
            mock.call(
                1,
                metrics=[
                    {
                        "expression": "COUNT(*)",
                        "metric_name": "cnt",
                        "metric_type": "count",
                        "verbose_name": "cnt",
                        "description": "",
                    },
                ],
            ),
        ],
    )


def test_sync_datasets_no_metrics(mocker: MockerFixture) -> None:
    """
    Test ``sync_datasets`` when no datasets exist yet.
    """
    client = mocker.MagicMock()
    client.get_datasets.return_value = []
    client.create_dataset.side_effect = [{"id": 1}, {"id": 2}, {"id": 3}]

    sync_datasets(
        client=client,
        models=models,
        metrics=[],
        database={"id": 1},
        disallow_edits=False,
        external_url_prefix="",
    )
    client.create_dataset.assert_has_calls(
        [
            mock.call(database=1, schema="public", table_name="messages_channels"),
        ],
    )
    client.update_dataset.assert_has_calls(
        [
            mock.call(
                1,
                description="",
                extra=json.dumps(
                    {
                        "unique_id": "model.superset_examples.messages_channels",
                        "depends_on": "ref('messages_channels')",
                        "certification": {"details": "This table is produced by dbt"},
                    },
                ),
                is_managed_externally=False,
                metrics=[],
            ),
        ],
    )


def test_sync_datasets_new_bq_error(mocker: MockerFixture) -> None:
    """
    Test ``sync_datasets`` when one of the sources is in a different BQ project.

    When that happens we can't add the dataset, since Superset can't read the metadata.
    """
    client = mocker.MagicMock()
    client.get_datasets.return_value = []
    client.create_dataset.side_effect = [
        Exception("Table is in a different project"),
    ]

    sync_datasets(
        client=client,
        models=models,
        metrics=metrics,
        database={"id": 1},
        disallow_edits=False,
        external_url_prefix="",
    )
    client.create_dataset.assert_has_calls(
        [
            mock.call(database=1, schema="public", table_name="messages_channels"),
        ],
    )
    client.update_dataset.assert_has_calls([])


def test_sync_datasets_existing(mocker: MockerFixture) -> None:
    """
    Test ``sync_datasets`` when datasets already exist.
    """
    client = mocker.MagicMock()
    client.get_datasets.side_effect = [[{"id": 1}], [{"id": 2}], [{"id": 3}]]

    sync_datasets(
        client=client,
        models=models,
        metrics=metrics,
        database={"id": 1},
        disallow_edits=False,
        external_url_prefix="",
    )
    client.create_dataset.assert_not_called()
    client.update_dataset.assert_has_calls(
        [
            mock.call(
                1,
                description="",
                extra=json.dumps(
                    {
                        "unique_id": "model.superset_examples.messages_channels",
                        "depends_on": "ref('messages_channels')",
                        "certification": {"details": "This table is produced by dbt"},
                    },
                ),
                is_managed_externally=False,
                metrics=[],
            ),
            mock.call(
                1,
                metrics=[
                    {
                        "expression": "COUNT(*)",
                        "metric_name": "cnt",
                        "metric_type": "count",
                        "verbose_name": "cnt",
                        "description": "",
                    },
                ],
            ),
        ],
    )


def test_sync_datasets_multiple_existing(mocker: MockerFixture) -> None:
    """
    Test ``sync_datasets`` when multiple datasets are found to exist.
    """
    client = mocker.MagicMock()
    client.get_datasets.return_value = [{"id": 1}, {"id": 2}]

    with pytest.raises(Exception) as excinfo:
        sync_datasets(
            client=client,
            models=models,
            metrics=metrics,
            database={"id": 1},
            disallow_edits=False,
            external_url_prefix="",
        )
    assert str(excinfo.value) == "More than one dataset found"


def test_sync_datasets_external_url(mocker: MockerFixture) -> None:
    """
    Test ``sync_datasets`` when passing external URL prefix.
    """
    client = mocker.MagicMock()
    client.get_datasets.side_effect = [[{"id": 1}], [{"id": 2}], [{"id": 3}]]

    sync_datasets(
        client=client,
        models=models,
        metrics=metrics,
        database={"id": 1},
        disallow_edits=False,
        external_url_prefix="https://dbt.example.org/",
    )
    client.create_dataset.assert_not_called()
    client.update_dataset.assert_has_calls(
        [
            mock.call(
                1,
                description="",
                extra=json.dumps(
                    {
                        "unique_id": "model.superset_examples.messages_channels",
                        "depends_on": "ref('messages_channels')",
                        "certification": {"details": "This table is produced by dbt"},
                    },
                ),
                is_managed_externally=False,
                metrics=[],
                external_url=(
                    "https://dbt.example.org/#!"
                    "/model/model.superset_examples.messages_channels"
                ),
            ),
            mock.call(
                1,
                metrics=[
                    {
                        "expression": "COUNT(*)",
                        "metric_name": "cnt",
                        "metric_type": "count",
                        "verbose_name": "cnt",
                        "description": "",
                    },
                ],
            ),
        ],
    )



"""
Tests for the import commands.
"""

# pylint: disable=invalid-name

import yaml
from click.testing import CliRunner
from pyfakefs.fake_filesystem import FakeFilesystem
from pytest_mock import MockerFixture

from preset_cli.cli.superset.main import superset_cli


def test_import_rls(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test the ``import_rls`` command.
    """
    mocker.patch("preset_cli.cli.superset.main.UsernamePasswordAuth")
    SupersetClient = mocker.patch("preset_cli.cli.superset.import_.SupersetClient")
    client = SupersetClient()
    rls = [
        {
            "clause": "client_id = 9",
            "description": "Rule description",
            "filter_type": "Regular",
            "group_key": "department",
            "name": "Rule name",
            "roles": ["Gamma"],
            "tables": ["main.test_table"],
        },
    ]
    fs.create_file("rls.yaml", contents=yaml.dump(rls))

    runner = CliRunner()
    result = runner.invoke(
        superset_cli,
        ["https://superset.example.org/", "import-rls"],
        catch_exceptions=False,
    )
    assert result.exit_code == 0

    client.import_rls.assert_called_with(rls[0])


def test_import_ownership(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test the ``import_ownership`` command.
    """
    mocker.patch("preset_cli.cli.superset.main.UsernamePasswordAuth")
    SupersetClient = mocker.patch("preset_cli.cli.superset.import_.SupersetClient")
    client = SupersetClient()
    ownership = {
        "dataset": [
            {
                "name": "test_table",
                "owners": ["admin@example.com"],
                "uuid": "e4e6a14b-c3e8-4fdf-a850-183ba6ce15e0",
            },
        ],
    }
    fs.create_file("ownership.yaml", contents=yaml.dump(ownership))

    runner = CliRunner()
    result = runner.invoke(
        superset_cli,
        ["https://superset.example.org/", "import-ownership"],
        catch_exceptions=False,
    )
    assert result.exit_code == 0

    client.import_ownership.assert_called_with("dataset", ownership["dataset"])



"""
Helped functions for authentication.
"""

from pathlib import Path
from typing import Union

import requests
import yaml
from appdirs import user_config_dir
from yarl import URL

CREDENTIALS_FILE = "credentials.yaml"


def get_access_token(baseurl: Union[str, URL], api_token: str, api_secret: str) -> str:
    """
    Fetch the JWT access token.
    """
    if isinstance(baseurl, str):
        baseurl = URL(baseurl)

    response = requests.post(
        baseurl / "api/v1/auth/",
        json={"name": api_token, "secret": api_secret},
        headers={"Content-Type": "application/json"},
    )
    payload = response.json()
    return payload["payload"]["access_token"]


def get_credentials_path() -> Path:
    """
    Return the system-dependent location of the credentials.
    """
    config_dir = Path(user_config_dir("preset-cli", "Preset"))
    return config_dir / CREDENTIALS_FILE


def store_credentials(
    api_token: str,
    api_secret: str,
    manager_url: URL,
    credentials_path: Path,
) -> None:
    """
    Store credentials.
    """
    credentials_path.parent.mkdir(parents=True, exist_ok=True)

    credentials = {
        "api_token": api_token,
        "api_secret": api_secret,
        "baseurl": str(manager_url),
    }

    while True:
        store = input(f"Store the credentials in {credentials_path}? [y/N] ")
        if store.strip().lower() == "y":
            with open(credentials_path, "w", encoding="utf-8") as output:
                yaml.safe_dump(credentials, output)
            credentials_path.chmod(0o600)
            break

        if store.strip().lower() in ("n", ""):
            break



"""
Commands for syncing metastores to and from Superset.
"""

import click

from preset_cli.cli.superset.sync.dbt.command import dbt_cloud, dbt_core
from preset_cli.cli.superset.sync.native.command import native


@click.group()
def sync() -> None:
    """
    Sync metadata between Superset and an external repository.
    """


sync.add_command(native)
sync.add_command(dbt_cloud)
sync.add_command(dbt_core)
# for backwards compatibility
sync.add_command(dbt_core, name="dbt")



"""
Tests for ``preset_cli.auth.lib``.
"""

from pathlib import Path

import yaml
from pyfakefs.fake_filesystem import FakeFilesystem
from pytest_mock import MockerFixture
from requests_mock.mocker import Mocker
from yarl import URL

from preset_cli.auth.lib import (
    get_access_token,
    get_credentials_path,
    store_credentials,
)


def test_get_access_token(requests_mock: Mocker) -> None:
    """
    Test ``get_access_token``.
    """
    requests_mock.post(
        "https://manage.app.preset.io/api/v1/auth/",
        json={"payload": {"access_token": "TOKEN"}},
    )

    access_token = get_access_token(
        URL("https://manage.app.preset.io/"),
        "API_TOKEN",
        "API_SECRET",
    )
    assert access_token == "TOKEN"

    access_token = get_access_token(
        "https://manage.app.preset.io/",
        "API_TOKEN",
        "API_SECRET",
    )
    assert access_token == "TOKEN"


def test_get_credentials_path(mocker: MockerFixture) -> None:
    """
    Test ``get_credentials_path``.
    """
    mocker.patch("preset_cli.auth.lib.user_config_dir", return_value="/path/to/config")
    assert get_credentials_path() == Path("/path/to/config/credentials.yaml")


# pylint: disable=unused-argument, invalid-name
def test_store_credentials(mocker: MockerFixture, fs: FakeFilesystem) -> None:
    """
    Test ``store_credentials``.
    """
    credentials_path = Path("/path/to/config/credentials.yaml")

    mocker.patch("preset_cli.auth.lib.input", side_effect=["invalid", "n"])
    store_credentials(
        "API_TOKEN",
        "API_SECRET",
        URL("https://manage.app.preset.io/"),
        credentials_path,
    )
    assert not credentials_path.exists()

    mocker.patch("preset_cli.auth.lib.input", return_value="y")
    store_credentials(
        "API_TOKEN",
        "API_SECRET",
        URL("https://manage.app.preset.io/"),
        credentials_path,
    )
    assert credentials_path.exists()
    with open(credentials_path, encoding="utf-8") as input_:
        contents = yaml.load(input_, Loader=yaml.SafeLoader)
    assert contents == {
        "api_secret": "API_SECRET",
        "api_token": "API_TOKEN",
        "baseurl": "https://manage.app.preset.io/",
    }



"""
Tests for ``preset_cli.lib``.
"""

import logging

import pytest
from pytest_mock import MockerFixture

from preset_cli.exceptions import ErrorLevel, SupersetError
from preset_cli.lib import remove_root, setup_logging, validate_response


def test_remove_root() -> None:
    """
    Test ``remove_root``.
    """
    assert remove_root("bundle/database/examples.yaml") == "database/examples.yaml"


def test_setup_logging() -> None:
    """
    Test ``setup_logging``.
    """
    setup_logging("debug")
    assert logging.root.level == logging.DEBUG

    with pytest.raises(ValueError) as excinfo:
        setup_logging("invalid")
    assert str(excinfo.value) == "Invalid log level: invalid"


def test_validate_response(mocker: MockerFixture) -> None:
    """
    Test ``validate_response``.
    """
    response = mocker.MagicMock()

    response.ok = True
    validate_response(response)

    # SIP-40 payload
    response.ok = False
    response.headers.get.return_value = "application/json"
    response.json.return_value = {
        "errors": [{"message": "Some message", "level": "error"}],
    }
    with pytest.raises(SupersetError) as excinfo:
        validate_response(response)
    assert excinfo.value.errors == [
        {"message": "Some message", "level": ErrorLevel.ERROR},
    ]

    # SIP-40-ish payload
    response.ok = False
    response.headers.get.return_value = "application/json"
    response.json.return_value = {"errors": {"message": "Some message"}}
    with pytest.raises(SupersetError) as excinfo:
        validate_response(response)
    assert excinfo.value.errors == [
        {
            "message": "Unknown error",
            "error_type": "UNKNOWN_ERROR",
            "level": ErrorLevel.ERROR,
            "extra": {"errors": {"message": "Some message"}},
        },
    ]

    # Old error
    response.ok = False
    response.headers.get.return_value = "application/json"
    response.json.return_value = {"message": "Some message"}
    with pytest.raises(SupersetError) as excinfo:
        validate_response(response)
    assert excinfo.value.errors == [
        {
            "message": "Unknown error",
            "error_type": "UNKNOWN_ERROR",
            "level": ErrorLevel.ERROR,
            "extra": {"message": "Some message"},
        },
    ]

    # Non JSON error
    response.ok = False
    response.headers.get.return_value = "text/plain"
    response.text = "An error occurred"
    with pytest.raises(SupersetError) as excinfo:
        validate_response(response)
    assert excinfo.value.errors == [
        {
            "message": "An error occurred",
            "error_type": "UNKNOWN_ERROR",
            "level": ErrorLevel.ERROR,
        },
    ]



"""
A simple client for running SQL queries (and more) against Superset:

    >>> from yarl import URL
    >>> from preset_cli.api.clients.superset import SupersetClient
    >>> from preset_cli.auth.main import UsernamePasswordAuth
    >>> url = URL("http://localhost:8088/")
    >>> auth = UsernamePasswordAuth(url, "admin", "admin")  # doctest: +SKIP
    >>> client = SupersetClient(url, auth)  # doctest: +SKIP
    >>> sql = "SELECT platform, rank FROM video_game_sales LIMIT 2"
    >>> print(client.run_query(database_id=1, sql=sql))  # doctest: +SKIP
      platform  rank
    0      Wii     1
    1      NES     2

Data is returned in a Pandas Dataframe.

"""

import json
import uuid
from datetime import datetime
from enum import IntEnum
from io import BytesIO
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Literal,
    Optional,
    Tuple,
    TypedDict,
    Union,
    cast,
)
from uuid import UUID, uuid4
from zipfile import ZipFile

import pandas as pd
import prison
import yaml
from bs4 import BeautifulSoup
from yarl import URL

from preset_cli import __version__
from preset_cli.api.operators import Equal, Operator
from preset_cli.auth.main import Auth
from preset_cli.lib import remove_root, validate_response

MAX_PAGE_SIZE = 100
MAX_IDS_IN_EXPORT = 100


class GenericDataType(IntEnum):
    """
    Generic database column type that fits both frontend and backend.
    """

    NUMERIC = 0
    STRING = 1
    TEMPORAL = 2
    BOOLEAN = 3


class AdhocMetricColumn(TypedDict, total=False):
    """
    Schema for an adhoc metric column.
    """

    column_name: Optional[str]
    description: Optional[str]
    expression: Optional[str]
    filterable: bool
    groupby: bool
    id: int
    is_dttm: bool
    python_date_format: Optional[str]
    type: str
    type_generic: GenericDataType
    verbose_name: Optional[str]


class MetricType(TypedDict):
    """
    Schema for an adhoc metric in the Chart API.
    """

    aggregate: Optional[str]
    column: Optional[AdhocMetricColumn]
    expressionType: Literal["SIMPLE", "SQL"]
    hasCustomLabel: Optional[bool]
    label: Optional[str]
    sqlExpression: Optional[str]
    isNew: bool
    optionName: str


def convert_to_adhoc_metric(expression: str) -> MetricType:
    """
    Convert an adhoc metric to an object.
    """
    return {
        "aggregate": None,
        "column": None,
        "expressionType": "SQL",
        "hasCustomLabel": False,
        "isNew": False,
        "label": expression,
        "optionName": f"metric_{uuid4()}",
        "sqlExpression": expression,
    }


class ColumnType(TypedDict):
    """
    Schema for an adhoc column in the Chart API.
    """

    label: str
    sqlExpression: str


def convert_to_adhoc_column(expression: str) -> ColumnType:
    """
    Convert an adhoc column to an object.
    """
    return {
        "label": expression,
        "sqlExpression": expression,
    }


def shortid() -> str:
    """
    Generate a short ID suited for a SQL Lab client ID.
    """
    return str(uuid.uuid4())[-12:]


def parse_html_array(value: str) -> List[str]:
    """
    Parse an array scraped from the HTML CRUD view.
    """
    value = value.strip()

    if value.startswith("[") and value.endswith("]"):
        parts = [part.strip() for part in value[1:-1].split(",")]
    else:
        parts = [part.strip() for part in value.split("\n")]

    return [part for part in parts if part.strip()]


class UserType(TypedDict):
    """
    Schema for a user.
    """

    id: int
    username: str
    role: List[str]
    first_name: str
    last_name: str
    email: str


class RuleType(TypedDict):
    """
    Schema for an RLS rule.
    """

    name: str
    description: Optional[str]
    filter_type: str
    tables: List[str]
    roles: List[str]
    group_key: str
    clause: str


class OwnershipType(TypedDict):
    """
    Schema for resource ownership.
    """

    name: str
    uuid: UUID
    owners: List[str]


class SupersetClient:  # pylint: disable=too-many-public-methods

    """
    A client for running queries against Superset.
    """

    def __init__(self, baseurl: Union[str, URL], auth: Auth):
        # convert to URL if necessary
        self.baseurl = URL(baseurl)
        self.auth = auth

    def run_query(
        self,
        database_id: int,
        sql: str,
        schema: Optional[str] = None,
        limit: int = 1000,
    ) -> pd.DataFrame:
        """
        Run a SQL query, returning a Pandas dataframe.
        """
        url = self.baseurl / "superset/sql_json/"
        data = {
            "client_id": shortid()[:10],
            "database_id": database_id,
            "json": True,
            "runAsync": False,
            "schema": schema,
            "sql": sql,
            "sql_editor_id": "1",
            "tab": "Untitled Query 2",
            "tmp_table_name": "",
            "select_as_cta": False,
            "ctas_method": "TABLE",
            "queryLimit": limit,
            "expand_data": True,
        }
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "User-Agent": f"Apache Superset Client ({__version__})",
            "Referer": str(self.baseurl),
        }
        headers.update(self.auth.get_headers())

        session = self.auth.get_session()
        response = session.post(url, json=data, headers=headers)
        validate_response(response)

        payload = response.json()

        return pd.DataFrame(payload["data"])

    def get_data(  # pylint: disable=too-many-locals, too-many-arguments
        self,
        dataset_id: int,
        metrics: List[str],
        columns: List[str],
        order_by: Optional[List[str]] = None,
        order_desc: bool = True,
        is_timeseries: bool = False,
        time_column: Optional[str] = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        granularity: Optional[str] = None,
        where: str = "",
        having: str = "",
        row_limit: int = 10000,
        force: bool = False,
    ) -> pd.DataFrame:
        """
        Run a dimensional query.
        """
        dataset = self.get_dataset(dataset_id)["result"]

        if time_column is None:
            time_columns = [
                column["column_name"]
                for column in dataset["columns"]
                if column["is_dttm"]
            ]
            if len(time_columns) > 1:
                options = ", ".join(time_columns)
                raise Exception(
                    f"Unable to determine time column, please pass `time_series` "
                    f"as one of: {options}",
                )
            time_column = time_columns[0]

        time_range = (
            "No filter"
            if start is None and end is None
            else f"{start or ''} : {end or ''}"
        )

        # convert adhoc metrics to a proper object, if needed
        metric_names = [metric["metric_name"] for metric in dataset["metrics"]]
        processed_metrics = [
            metric if metric in metric_names else convert_to_adhoc_metric(metric)
            for metric in metrics
        ]

        # same for columns
        column_names = [column["column_name"] for column in dataset["columns"]]
        processed_columns = [
            column if column in column_names else convert_to_adhoc_column(column)
            for column in columns
        ]

        # and order bys
        processed_orderbys = [
            (orderby, not order_desc)
            if orderby in metric_names
            else (convert_to_adhoc_metric(orderby), not order_desc)
            for orderby in (order_by or [])
        ]

        url = self.baseurl / "api/v1/chart/data"
        data: Dict[str, Any] = {
            "datasource": {"id": dataset_id, "type": "table"},
            "force": force,
            "queries": [
                {
                    "annotation_layers": [],
                    "applied_time_extras": {},
                    "columns": processed_columns,
                    "custom_form_data": {},
                    "custom_params": {},
                    "extras": {"having": having, "having_druid": [], "where": where},
                    "filters": [],
                    "is_timeseries": is_timeseries,
                    "metrics": processed_metrics,
                    "order_desc": order_desc,
                    "orderby": processed_orderbys,
                    "row_limit": row_limit,
                    "time_range": time_range,
                    "timeseries_limit": 0,
                    "url_params": {},
                },
            ],
            "result_format": "json",
            "result_type": "full",
        }

        if is_timeseries:
            data["queries"][0]["granularity"] = time_column
            data["queries"][0]["extras"]["time_grain_sqla"] = granularity

        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "User-Agent": f"Apache Superset Client ({__version__})",
            "Referer": str(self.baseurl),
        }
        headers.update(self.auth.get_headers())

        session = self.auth.get_session()
        response = session.post(url, json=data, headers=headers)
        validate_response(response)

        payload = response.json()

        return pd.DataFrame(payload["result"][0]["data"])

    def get_resource(self, resource_name: str, resource_id: int) -> Any:
        """
        Return a single resource.
        """
        url = self.baseurl / "api/v1" / resource_name / str(resource_id)

        session = self.auth.get_session()
        headers = self.auth.get_headers()
        headers["Referer"] = str(self.baseurl)
        response = session.get(url, headers=headers)
        validate_response(response)

        resource = response.json()

        return resource

    def get_resources(self, resource_name: str, **kwargs: Any) -> List[Any]:
        """
        Return one or more of a resource, possibly filtered.
        """
        resources = []
        operations = {
            k: v if isinstance(v, Operator) else Equal(v) for k, v in kwargs.items()
        }

        # paginate endpoint until no results are returned
        page = 0
        while True:
            query = prison.dumps(
                {
                    "filters": [
                        dict(col=col, opr=value.operator, value=value.value)
                        for col, value in operations.items()
                    ],
                    "order_column": "changed_on_delta_humanized",
                    "order_direction": "desc",
                    "page": page,
                    "page_size": MAX_PAGE_SIZE,
                },
            )
            url = self.baseurl / "api/v1" / resource_name / "" % {"q": query}

            session = self.auth.get_session()
            headers = self.auth.get_headers()
            headers["Referer"] = str(self.baseurl)
            response = session.get(url, headers=headers)
            validate_response(response)

            payload = response.json()

            if not payload["result"]:
                break

            resources.extend(payload["result"])
            page += 1

        return resources

    def create_resource(self, resource_name: str, **kwargs: Any) -> Any:
        """
        Create a resource.
        """
        url = self.baseurl / "api/v1" / resource_name / ""

        session = self.auth.get_session()
        headers = self.auth.get_headers()
        headers["Referer"] = str(self.baseurl)
        response = session.post(url, json=kwargs, headers=headers)
        validate_response(response)

        resource = response.json()

        return resource

    def update_resource(
        self,
        resource_name: str,
        resource_id: int,
        query_args: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> Any:
        """
        Update a resource.
        """
        url = self.baseurl / "api/v1" / resource_name / str(resource_id)
        if query_args:
            url %= query_args

        session = self.auth.get_session()
        headers = self.auth.get_headers()
        headers["Referer"] = str(self.baseurl)
        response = session.put(url, json=kwargs, headers=headers)
        validate_response(response)

        resource = response.json()

        return resource

    def get_database(self, database_id: int) -> Any:
        """
        Return a single database.
        """
        return self.get_resource("database", database_id)

    def get_databases(self, **kwargs: str) -> List[Any]:
        """
        Return databases, possibly filtered.
        """
        return self.get_resources("database", **kwargs)

    def create_database(self, **kwargs: Any) -> Any:
        """
        Create a database.
        """
        return self.create_resource("database", **kwargs)

    def update_database(self, database_id: int, **kwargs: Any) -> Any:
        """
        Update a database.
        """
        query_args = {"override_columns": "true"}
        return self.update_resource("database", database_id, query_args, **kwargs)

    def get_dataset(self, dataset_id: int) -> Any:
        """
        Return a single dataset.
        """
        return self.get_resource("dataset", dataset_id)

    def get_datasets(self, **kwargs: str) -> List[Any]:
        """
        Return datasets, possibly filtered.
        """
        return self.get_resources("dataset", **kwargs)

    def create_dataset(self, **kwargs: Any) -> Any:
        """
        Create a dataset.
        """
        return self.create_resource("dataset", **kwargs)

    def update_dataset(self, dataset_id: int, **kwargs: Any) -> Any:
        """
        Update a dataset.
        """
        return self.update_resource("dataset", dataset_id, **kwargs)

    def get_chart(self, chart_id: int) -> Any:
        """
        Return a single chart.
        """
        return self.get_resource("chart", chart_id)

    def get_charts(self, **kwargs: str) -> List[Any]:
        """
        Return charts, possibly filtered.
        """
        return self.get_resources("chart", **kwargs)

    def get_dashboard(self, dashboard_id: int) -> Any:
        """
        Return a single dashboard.
        """
        return self.get_resource("dashboard", dashboard_id)

    def get_dashboards(self, **kwargs: str) -> List[Any]:
        """
        Return dashboards, possibly filtered.
        """
        return self.get_resources("dashboard", **kwargs)

    def create_dashboard(self, **kwargs: Any) -> Any:
        """
        Create a dashboard.
        """
        return self.create_resource("dashboard", **kwargs)

    def update_dashboard(self, dashboard_id: int, **kwargs: Any) -> Any:
        """
        Update a dashboard.
        """
        return self.update_resource("dashboard", dashboard_id, **kwargs)

    def export_zip(self, resource_name: str, ids: List[int]) -> BytesIO:
        """
        Export one or more of a resource.
        """
        session = self.auth.get_session()
        headers = self.auth.get_headers()
        headers["Referer"] = str(self.baseurl)
        url = self.baseurl / "api/v1" / resource_name / "export/"

        buf = BytesIO()
        with ZipFile(buf, "w") as bundle:
            while ids:
                page, ids = ids[:MAX_IDS_IN_EXPORT], ids[MAX_IDS_IN_EXPORT:]
                params = {"q": prison.dumps(page)}
                response = session.get(url, params=params, headers=headers)
                validate_response(response)

                # write files from response to main ZIP bundle
                with ZipFile(BytesIO(response.content)) as subset:
                    for name in subset.namelist():
                        bundle.writestr(name, subset.read(name))

        buf.seek(0)

        return buf

    def get_uuids(self, resource_name: str) -> Dict[int, UUID]:
        """
        Get UUID of a list of resources.

        Still method is very inneficient, but it's the only way to get the mapping
        between IDs and UUIDs in older versions of Superset.
        """
        session = self.auth.get_session()
        headers = self.auth.get_headers()
        headers["Referer"] = str(self.baseurl)
        url = self.baseurl / "api/v1" / resource_name / "export/"

        uuids: Dict[int, UUID] = {}
        for resource in self.get_resources(resource_name):
            id_ = resource["id"]
            params = {"q": prison.dumps([id_])}
            response = session.get(url, params=params, headers=headers)

            with ZipFile(BytesIO(response.content)) as export:
                for name in export.namelist():
                    config = yaml.load(export.read(name), Loader=yaml.SafeLoader)
                    name = remove_root(name)
                    if name.startswith(resource_name):
                        uuids[id_] = UUID(config["uuid"])

        return uuids

    def import_zip(
        self,
        resource_name: str,
        data: BytesIO,
        overwrite: bool = False,
    ) -> bool:
        """
        Import a ZIP bundle.
        """
        url = self.baseurl / "api/v1" / resource_name / "import/"

        session = self.auth.get_session()
        headers = self.auth.get_headers()
        headers["Referer"] = str(self.baseurl)
        headers["Accept"] = "application/json"
        response = session.post(
            url,
            files=dict(formData=data),
            data=dict(overwrite=json.dumps(overwrite)),
            headers=headers,
        )
        validate_response(response)

        payload = response.json()

        return payload["message"] == "OK"

    def export_users(self) -> Iterator[UserType]:
        """
        Return all users.

        Since this is not exposed via an API we need to crawl the CRUD page.
        """
        session = self.auth.get_session()
        headers = self.auth.get_headers()
        headers["Referer"] = str(self.baseurl)

        page = 0
        while True:
            params = {
                "psize_UserDBModelView": MAX_PAGE_SIZE,
                "page_UserDBModelView": page,
            }
            url = self.baseurl / "users/list/"
            page += 1

            response = session.get(url, params=params, headers=headers)
            soup = BeautifulSoup(response.text, features="html.parser")
            table = soup.find_all("table")[1]
            trs = table.find_all("tr")
            if len(trs) == 1:
                break

            for tr in trs[1:]:  # pylint: disable=invalid-name
                tds = tr.find_all("td")
                yield {
                    "id": int(tds[0].find("a").attrs["href"].split("/")[-1]),
                    "first_name": tds[1].text,
                    "last_name": tds[2].text,
                    "username": tds[3].text,
                    "email": tds[4].text,
                    "role": parse_html_array(tds[6].text.strip()),
                }

    def export_rls(self) -> Iterator[RuleType]:
        """
        Return all RLS rules.
        """
        session = self.auth.get_session()
        headers = self.auth.get_headers()
        headers["Referer"] = str(self.baseurl)

        page = 0
        while True:
            params = {
                "psize_RowLevelSecurityFiltersModelView": MAX_PAGE_SIZE,
                "page_RowLevelSecurityFiltersModelView": page,
            }
            url = self.baseurl / "rowlevelsecurityfiltersmodelview/list/"
            page += 1

            response = session.get(url, params=params, headers=headers)
            soup = BeautifulSoup(response.text, features="html.parser")
            table = soup.find_all("table")[1]
            trs = table.find_all("tr")
            if len(trs) == 1:
                break

            for tr in trs[1:]:  # pylint: disable=invalid-name
                tds = tr.find_all("td")

                # extract the ID to fetch each RLS in a separate request, since the list
                # view doesn't have all the columns we need
                rule_id = int(tds[0].find("input").attrs["id"])
                rule_url = (
                    self.baseurl
                    / "rowlevelsecurityfiltersmodelview/show"
                    / str(rule_id)
                )

                response = session.get(rule_url, headers=headers)
                soup = BeautifulSoup(response.text, features="html.parser")
                table = soup.find("table")
                keys: List[Tuple[str, Callable[[Any], Any]]] = [
                    ("name", str),
                    ("description", str),
                    ("filter_type", str),
                    ("tables", parse_html_array),
                    ("roles", parse_html_array),
                    ("group_key", str),
                    ("clause", str),
                ]
                yield cast(
                    RuleType,
                    {
                        key: parse(tr.find("td").text.strip())
                        for (key, parse), tr in zip(keys, table.find_all("tr"))
                    },
                )

    def import_rls(self, rls: RuleType) -> None:  # pylint: disable=too-many-locals
        """
        Import a given RLS rule.
        """
        session = self.auth.get_session()
        headers = self.auth.get_headers()
        headers["Referer"] = str(self.baseurl)

        table_ids: List[int] = []
        for table in rls["tables"]:
            if "." in table:
                schema, table_name = table.split(".", 1)
                datasets = self.get_datasets(schema=schema, table_name=table_name)
            else:
                datasets = self.get_datasets(table_name=table)

            if not datasets:
                raise Exception(f"Cannot find table: {table}")
            if len(datasets) > 1:
                raise Exception(f"More than one table found: {table}")
            table_ids.append(datasets[0]["id"])

        role_ids: List[int] = []
        for role in rls["roles"]:
            params = {"_flt_0_name": role}
            url = self.baseurl / "roles/list/"
            response = session.get(url, params=params, headers=headers)
            soup = BeautifulSoup(response.text, features="html.parser")
            trs = soup.find_all("table")[1].find_all("tr")
            if len(trs) == 1:
                raise Exception(f"Cannot find role: {role}")
            if len(trs) > 2:
                raise Exception(f"More than one role found: {role}")
            td = trs[1].find("td")  # pylint: disable=invalid-name
            if td.find("a"):
                id_ = int(td.find("a").attrs["href"].split("/")[-1])
            else:
                id_ = int(td.find("input").attrs["id"])
            role_ids.append(id_)

        url = self.baseurl / "rowlevelsecurityfiltersmodelview/add"
        response = session.post(
            url,
            data={
                # "csrf_token": "needed?",
                "name": rls["name"],
                "description": rls["description"],
                "filter_type": rls["filter_type"],
                "tables": table_ids,
                "roles": role_ids,
                "group_key": rls["group_key"],
                "clause": rls["clause"],
            },
            headers=headers,
        )
        validate_response(response)

    def export_ownership(self, resource_name: str) -> Iterator[OwnershipType]:
        """
        Return information about resource ownership.
        """
        emails = {user["id"]: user["email"] for user in self.export_users()}
        uuids = self.get_uuids(resource_name)
        name_key = {
            "dataset": "table_name",
            "chart": "slice_name",
            "dashboard": "dashboard_title",
        }[resource_name]

        for resource in self.get_resources(resource_name):
            yield {
                "name": resource[name_key],
                "uuid": uuids[resource["id"]],
                "owners": [emails[owner["id"]] for owner in resource.get("owners", [])],
            }

    def import_ownership(
        self,
        resource_name: str,
        ownership: List[Dict[str, Any]],
    ) -> None:
        """
        Import ownership on resources.
        """
        user_ids = {user["email"]: user["id"] for user in self.export_users()}
        resource_ids = {str(v): k for k, v in self.get_uuids(resource_name).items()}

        for item in ownership:
            resource_id = resource_ids[item["uuid"]]
            owner_ids = [user_ids[email] for email in item["owners"]]
            self.update_resource(resource_name, resource_id, owners=owner_ids)



"""
Tests for the dbt client.
"""

# pylint: disable=missing-class-docstring, invalid-name, line-too-long

import datetime
from enum import Enum

import pytest
from marshmallow import fields
from pytest_mock import MockerFixture
from requests_mock.mocker import Mocker

from preset_cli import __version__
from preset_cli.api.clients.dbt import DBTClient, PostelEnumField
from preset_cli.auth.main import Auth


def test_postel_enum_field() -> None:
    """
    Test ``PostelEnumField``.
    """

    class StrEnum(str, Enum):
        A = "A"

    class IntEnum(int, Enum):
        A = 1

    class RawEnum(Enum):
        A = "A"
        B = 1

    assert isinstance(PostelEnumField(StrEnum, allow_none=True), fields.String)
    assert isinstance(PostelEnumField(IntEnum, allow_none=False), fields.Integer)
    assert isinstance(PostelEnumField(RawEnum), fields.Raw)


def test_dbt_client_execute(mocker: MockerFixture) -> None:
    """
    Test the ``execute`` method.
    """
    GraphqlClient = mocker.patch("preset_cli.api.clients.dbt.GraphqlClient")
    auth = Auth()
    client = DBTClient(auth)
    query = """
        query ($jobId: Int!) {
            models(jobId: $jobId) {
                uniqueId
                name
                database
                schema
                description
                meta
            }
        }
    """
    client.execute(query, jobId=1)
    GraphqlClient().execute.assert_called_with(
        query=query,
        variables={"jobId": 1},
        headers={
            "User-Agent": "Preset CLI",
            "X-Client-Version": __version__,
        },
    )


def test_dbt_client_get_accounts(requests_mock: Mocker) -> None:
    """
    Test the ``get_accounts`` method.
    """
    requests_mock.get(
        "https://cloud.getdbt.com/api/v2/accounts/",
        json={
            "status": {
                "code": 200,
                "is_success": True,
                "user_message": "Success!",
                "developer_message": "",
            },
            "data": [
                {
                    "docs_job_id": None,
                    "freshness_job_id": None,
                    "lock_reason": "[SYSTEM] Trial was expired as of 2022-06-24.",
                    "unlock_if_subscription_renewed": True,
                    "read_only_seats": 50,
                    "id": 72449,
                    "name": "Preset (Partner)",
                    "state": 1,
                    "plan": "free",
                    "pending_cancel": False,
                    "run_slots": 5,
                    "developer_seats": 50,
                    "queue_limit": 50,
                    "pod_memory_request_mebibytes": 600,
                    "run_duration_limit_seconds": 86400,
                    "enterprise_authentication_method": None,
                    "enterprise_login_slug": None,
                    "enterprise_unique_identifier": None,
                    "billing_email_address": None,
                    "locked": False,
                    "develop_file_system": True,
                    "unlocked_at": "2022-07-20T19:03:31.337718+00:00",
                    "created_at": "2022-06-09T21:42:07.004082+00:00",
                    "updated_at": "2022-07-20T19:03:31.341319+00:00",
                    "starter_repo_url": None,
                    "sso_reauth": True,
                    "force_sso": True,
                    "git_auth_level": "team",
                    "identifier": "act_2AMECVyWdvvg34fJWwVMWGtuTwq",
                    "docs_job": None,
                    "freshness_job": None,
                    "enterprise_login_url": "https://cloud.getdbt.com/enterprise-login/None/",
                },
            ],
            "extra": {
                "filters": {"pk": 72449},
                "order_by": None,
                "pagination": {"count": 1, "total_count": 1},
            },
        },
    )
    auth = Auth()
    client = DBTClient(auth)
    assert client.get_accounts() == [
        {
            "run_slots": 5,
            "created_at": datetime.datetime(
                2022,
                6,
                9,
                21,
                42,
                7,
                4082,
                tzinfo=datetime.timezone(datetime.timedelta(0), "+0000"),
            ),
            "name": "Preset (Partner)",
            "state": 1,
            "updated_at": datetime.datetime(
                2022,
                7,
                20,
                19,
                3,
                31,
                341319,
                tzinfo=datetime.timezone(datetime.timedelta(0), "+0000"),
            ),
            "read_only_seats": 50,
            "id": 72449,
            "developer_seats": 50,
            "plan": "free",
            "pending_cancel": False,
            "enterprise_login_url": "https://cloud.getdbt.com/enterprise-login/None/",
            "enterprise_authentication_method": None,
            "starter_repo_url": None,
            "force_sso": True,
            "develop_file_system": True,
            "lock_reason": "[SYSTEM] Trial was expired as of 2022-06-24.",
            "identifier": "act_2AMECVyWdvvg34fJWwVMWGtuTwq",
            "billing_email_address": None,
            "docs_job": None,
            "freshness_job": None,
            "enterprise_login_slug": None,
            "freshness_job_id": None,
            "locked": False,
            "unlocked_at": "2022-07-20T19:03:31.337718+00:00",
            "docs_job_id": None,
            "unlock_if_subscription_renewed": True,
            "run_duration_limit_seconds": 86400,
            "sso_reauth": True,
            "git_auth_level": "team",
            "enterprise_unique_identifier": None,
            "queue_limit": 50,
            "pod_memory_request_mebibytes": 600,
        },
    ]


def test_dbt_client_get_projects(requests_mock: Mocker) -> None:
    """
    Test the ``get_projects`` method.
    """
    requests_mock.get(
        "https://cloud.getdbt.com/api/v2/accounts/72449/projects/",
        json={
            "status": {
                "code": 200,
                "is_success": True,
                "user_message": "Success!",
                "developer_message": "",
            },
            "data": [
                {
                    "name": "Analytics",
                    "account_id": 72449,
                    "repository_id": 85562,
                    "connection_id": 79280,
                    "id": 113498,
                    "created_at": "2022-06-09 21:42:07.072884+00:00",
                    "updated_at": "2022-07-25 21:53:45.321518+00:00",
                    "skipped_setup": False,
                    "state": 1,
                    "dbt_project_subdirectory": None,
                    "connection": {
                        "id": 79280,
                        "account_id": 72449,
                        "project_id": 113498,
                        "name": "Bigquery",
                        "type": "bigquery",
                        "created_by_id": 89392,
                        "created_by_service_token_id": None,
                        "details": {
                            "project_id": "dbt-tutorial-347100",
                            "timeout_seconds": 300,
                            "private_key_id": "39f4cf7ffd957f8eaaf41e1009401f9c3fb039d6",
                            "client_email": "dbt-user@dbt-tutorial-347100.iam.gserviceaccount.com",
                            "client_id": "106702535903205632050",
                            "auth_uri": "https://accounts.google.com/o/oauth2/auth",
                            "token_uri": "https://oauth2.googleapis.com/token",
                            "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
                            "client_x509_cert_url": "https://www.googleapis.com/robot/v1/metadata/x509/dbt-user%40dbt-tutorial-347100.iam.gserviceaccount.com",
                            "priority": "interactive",
                            "retries": 1,
                            "scopes": [
                                "https://www.googleapis.com/auth/bigquery",
                                "https://www.googleapis.com/auth/cloud-platform",
                                "https://www.googleapis.com/auth/drive",
                            ],
                            "location": None,
                            "maximum_bytes_billed": None,
                            "execution_project": None,
                            "impersonate_service_account": None,
                            "job_retry_deadline_seconds": None,
                            "job_creation_timeout_seconds": None,
                            "is_configured_for_oauth": False,
                        },
                        "state": 1,
                        "created_at": "2022-07-25 21:52:26.669562+00:00",
                        "updated_at": "2022-07-25 21:52:26.669581+00:00",
                    },
                    "repository": {
                        "id": 85562,
                        "account_id": 72449,
                        "project_id": 113498,
                        "full_name": "shreesham/preset-dbt-demo",
                        "remote_url": "git://github.com/shreesham/preset-dbt-demo.git",
                        "remote_backend": "github",
                        "git_clone_strategy": "github_app",
                        "deploy_key_id": 84338,
                        "repository_credentials_id": None,
                        "github_installation_id": 27417600,
                        "pull_request_url_template": "https://github.com/shreesham/preset-dbt-demo/compare/{{destination}}...{{source}}",
                        "state": 1,
                        "created_at": "2022-07-25 21:53:44.205393+00:00",
                        "updated_at": "2022-07-25 21:53:44.205413+00:00",
                        "deploy_key": {
                            "id": 84338,
                            "account_id": 72449,
                            "state": 1,
                            "public_key": "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAACAQCsuzl8uH0z0395SmYnlNuyeOCUqF7UhDVrIpUe+Vsyjin35cfnPop19k9l7nr+AZma+p2lf4NW9K1ugErB7hr5VUsSEoNjFk292z6dIhQVSLXrMJr3Kc0PixgXuDMgHUPQTU3T8zWeUmHQsslK0jTiRT/zH5ZdLc36AW9U2X9KriNHAfDHpx8b6eAgx2Prsvn/rnVqyolGlk9QBAPFqPzWbw231ORvWozUd94gklxJpWzIOvhZqVluuPs32FLquPfdkkRDjoVoLR1TQtriBBjaKsx6TK/frwtiOvrmbRm9l2ufXmHlahgTusbPru99z1rEwGlrqOI+DVo5VqM0ksavSz4auTJmzoIDhVXPgF2snwdnpfDeXiW9N/GQ90KGHGienBXnMeDCu+XcQrn+gUDimJfnRsSPSgU27eG/zWl6dEerlprV60g5WAViJ/MF1hBHZW+dpGJAipobvhA9NOA53xOHXYdVNn9uk9092/lQeiIBowhr8PKYezKMIf8z8cPKqL0BtUpBbYR/UpnmeWx9CmFbx9GxITiIWx+qOxZAzrLHwHrTfu1Rb20dXcE3XHSgmRFyyitOHn2+JN+rdZBT3XNePM++vygImnHoDCkJ/q11xmJWOEshsKESz0JrxGIkvXOJ3/2/Gw2Iujs17dFtXJADBzkNyuqGj4mD1Z/8gQ==",
                        },
                        "github_repo": "shreesham/preset-dbt-demo",
                        "name": "preset-dbt-demo",
                        "git_provider_id": 33120,
                        "gitlab": None,
                        "git_provider": None,
                    },
                    "group_permissions": [],
                    "docs_job_id": None,
                    "freshness_job_id": None,
                    "docs_job": None,
                    "freshness_job": None,
                },
                {
                    "name": "jaffle_shop",
                    "account_id": 72449,
                    "repository_id": 85563,
                    "connection_id": 79281,
                    "id": 134905,
                    "created_at": "2022-07-25 21:56:03.146661+00:00",
                    "updated_at": "2022-08-01 23:49:03.645162+00:00",
                    "skipped_setup": False,
                    "state": 1,
                    "dbt_project_subdirectory": "jaffle_shop",
                    "connection": {
                        "id": 79281,
                        "account_id": 72449,
                        "project_id": 134905,
                        "name": "Bigquery",
                        "type": "bigquery",
                        "created_by_id": 89392,
                        "created_by_service_token_id": None,
                        "details": {
                            "project_id": "dbt-tutorial-347100",
                            "timeout_seconds": 300,
                            "private_key_id": "39f4cf7ffd957f8eaaf41e1009401f9c3fb039d6",
                            "client_email": "dbt-user@dbt-tutorial-347100.iam.gserviceaccount.com",
                            "client_id": "106702535903205632050",
                            "auth_uri": "https://accounts.google.com/o/oauth2/auth",
                            "token_uri": "https://oauth2.googleapis.com/token",
                            "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
                            "client_x509_cert_url": "https://www.googleapis.com/robot/v1/metadata/x509/dbt-user%40dbt-tutorial-347100.iam.gserviceaccount.com",
                            "priority": "interactive",
                            "retries": 1,
                            "scopes": [
                                "https://www.googleapis.com/auth/bigquery",
                                "https://www.googleapis.com/auth/cloud-platform",
                                "https://www.googleapis.com/auth/drive",
                            ],
                            "location": None,
                            "maximum_bytes_billed": None,
                            "execution_project": None,
                            "impersonate_service_account": None,
                            "job_retry_deadline_seconds": None,
                            "job_creation_timeout_seconds": None,
                            "is_configured_for_oauth": False,
                        },
                        "state": 1,
                        "created_at": "2022-07-25 21:57:25.938478+00:00",
                        "updated_at": "2022-07-25 21:57:25.938494+00:00",
                    },
                    "repository": {
                        "id": 85563,
                        "account_id": 72449,
                        "project_id": 134905,
                        "full_name": "preset-io/dbt-integration-blog-post",
                        "remote_url": "git://github.com/preset-io/dbt-integration-blog-post.git",
                        "remote_backend": "github",
                        "git_clone_strategy": "github_app",
                        "deploy_key_id": 84339,
                        "repository_credentials_id": None,
                        "github_installation_id": 27693839,
                        "pull_request_url_template": "https://github.com/preset-io/dbt-integration-blog-post/compare/{{destination}}...{{source}}",
                        "state": 1,
                        "created_at": "2022-07-25 21:57:36.748283+00:00",
                        "updated_at": "2022-07-25 21:57:36.748299+00:00",
                        "deploy_key": {
                            "id": 84339,
                            "account_id": 72449,
                            "state": 1,
                            "public_key": "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAACAQCQ6NWTsb48rEVFc7/Ihsz1XQ6dFa0TQEntuDu6uOPUxBVadup83vhKNqI9xRuOnDDEm/YqhcUZqx6OFB0BFmxSeEoRwTkDkURfBKh1nQ1kSbq0xbKwoRXyvwyfqs1HeTEtSZuPcKelMOWK3R5evCE3ZcKvLQw6oHf32JhHThuaOw/AG5OtvSLlWSxlQ6b7hW64K2kdsf+KODnD9+6PhEMludcBtIwTcZQdhHbyz/UMo3fGas59+LsxhBzjVyF1jmi3XrKLHJ+5oTlvz+Jr2EI2Srvftcciy/DvyGvG/biqrG1KjrCYoqzwjCFhBy4jJBU421QTjBtYLABVJC9hEtlsSFCFMJMubUhKELo3HhX6vbn5E8TF+UCP7wc7JEACyICzxeROk2QAb0aV7gmL/Zrc6mVNou5Cb5mVuv5NIUBjpx7m69rsHRR+lb4H41gZCeeh7dQmV08rZk0L47n8Clcuc55v99o8H7ekk+I9dOhHpeKvbW90XMGtDnZZLfw+o2wG/9S4js1b142nvFP55TfAb6+wOa4W+ZidaH5E3Mp41uk2wgLE2CsRVjAP5vM3OFTU6iyDY/obOMERNEhThKUS83ixR67s5haoWLRy+q36SvLn1oYQYlJ0UZp3SxZUnUTV1m20L2oQ4jAhnr5j3zzsNWCZLW4SxgdOq7Mi16Bqow==",
                        },
                        "github_repo": "preset-io/dbt-integration-blog-post",
                        "name": "dbt-integration-blog-post",
                        "git_provider_id": 33120,
                        "gitlab": None,
                        "git_provider": None,
                    },
                    "group_permissions": [],
                    "docs_job_id": None,
                    "freshness_job_id": None,
                    "docs_job": None,
                    "freshness_job": None,
                },
            ],
            "extra": {
                "filters": {"account_id": 72449, "limit": 100, "offset": 0},
                "order_by": "id",
                "pagination": {"count": 2, "total_count": 2},
            },
        },
    )
    auth = Auth()
    client = DBTClient(auth)
    assert client.get_projects(72449) == [
        {
            "connection": {
                "details": {
                    "project_id": "dbt-tutorial-347100",
                    "timeout_seconds": 300,
                    "private_key_id": "39f4cf7ffd957f8eaaf41e1009401f9c3fb039d6",
                    "client_email": "dbt-user@dbt-tutorial-347100.iam.gserviceaccount.com",
                    "client_id": "106702535903205632050",
                    "auth_uri": "https://accounts.google.com/o/oauth2/auth",
                    "token_uri": "https://oauth2.googleapis.com/token",
                    "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
                    "client_x509_cert_url": "https://www.googleapis.com/robot/v1/metadata/x509/dbt-user%40dbt-tutorial-347100.iam.gserviceaccount.com",
                    "priority": "interactive",
                    "retries": 1,
                    "scopes": [
                        "https://www.googleapis.com/auth/bigquery",
                        "https://www.googleapis.com/auth/cloud-platform",
                        "https://www.googleapis.com/auth/drive",
                    ],
                    "location": None,
                    "maximum_bytes_billed": None,
                    "execution_project": None,
                    "impersonate_service_account": None,
                    "job_retry_deadline_seconds": None,
                    "job_creation_timeout_seconds": None,
                    "is_configured_for_oauth": False,
                },
                "type": "bigquery",
                "id": 79280,
                "account_id": 72449,
                "state": 1,
                "updated_at": datetime.datetime(
                    2022,
                    7,
                    25,
                    21,
                    52,
                    26,
                    669581,
                    tzinfo=datetime.timezone(datetime.timedelta(0), "+0000"),
                ),
                "created_by_id": 89392,
                "project_id": 113498,
                "created_by_service_token_id": None,
                "created_at": datetime.datetime(
                    2022,
                    7,
                    25,
                    21,
                    52,
                    26,
                    669562,
                    tzinfo=datetime.timezone(datetime.timedelta(0), "+0000"),
                ),
                "name": "Bigquery",
            },
            "docs_job": None,
            "freshness_job_id": None,
            "freshness_job": None,
            "id": 113498,
            "connection_id": 79280,
            "account_id": 72449,
            "repository": {
                "gitlab": None,
                "deploy_key_id": 84338,
                "account_id": 72449,
                "repository_credentials_id": None,
                "git_provider_id": 33120,
                "git_clone_strategy": "github_app",
                "deploy_key": {
                    "id": 84338,
                    "account_id": 72449,
                    "state": 1,
                    "public_key": "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAACAQCsuzl8uH0z0395SmYnlNuyeOCUqF7UhDVrIpUe+Vsyjin35cfnPop19k9l7nr+AZma+p2lf4NW9K1ugErB7hr5VUsSEoNjFk292z6dIhQVSLXrMJr3Kc0PixgXuDMgHUPQTU3T8zWeUmHQsslK0jTiRT/zH5ZdLc36AW9U2X9KriNHAfDHpx8b6eAgx2Prsvn/rnVqyolGlk9QBAPFqPzWbw231ORvWozUd94gklxJpWzIOvhZqVluuPs32FLquPfdkkRDjoVoLR1TQtriBBjaKsx6TK/frwtiOvrmbRm9l2ufXmHlahgTusbPru99z1rEwGlrqOI+DVo5VqM0ksavSz4auTJmzoIDhVXPgF2snwdnpfDeXiW9N/GQ90KGHGienBXnMeDCu+XcQrn+gUDimJfnRsSPSgU27eG/zWl6dEerlprV60g5WAViJ/MF1hBHZW+dpGJAipobvhA9NOA53xOHXYdVNn9uk9092/lQeiIBowhr8PKYezKMIf8z8cPKqL0BtUpBbYR/UpnmeWx9CmFbx9GxITiIWx+qOxZAzrLHwHrTfu1Rb20dXcE3XHSgmRFyyitOHn2+JN+rdZBT3XNePM++vygImnHoDCkJ/q11xmJWOEshsKESz0JrxGIkvXOJ3/2/Gw2Iujs17dFtXJADBzkNyuqGj4mD1Z/8gQ==",
                },
                "full_name": "shreesham/preset-dbt-demo",
                "updated_at": datetime.datetime(
                    2022,
                    7,
                    25,
                    21,
                    53,
                    44,
                    205413,
                    tzinfo=datetime.timezone(datetime.timedelta(0), "+0000"),
                ),
                "github_repo": "shreesham/preset-dbt-demo",
                "pull_request_url_template": "https://github.com/shreesham/preset-dbt-demo/compare/{{destination}}...{{source}}",
                "created_at": datetime.datetime(
                    2022,
                    7,
                    25,
                    21,
                    53,
                    44,
                    205393,
                    tzinfo=datetime.timezone(datetime.timedelta(0), "+0000"),
                ),
                "git_provider": None,
                "id": 85562,
                "remote_url": "git://github.com/shreesham/preset-dbt-demo.git",
                "state": 1,
                "github_installation_id": 27417600,
                "project_id": 113498,
                "remote_backend": "github",
                "name": "preset-dbt-demo",
            },
            "state": 1,
            "updated_at": datetime.datetime(
                2022,
                7,
                25,
                21,
                53,
                45,
                321518,
                tzinfo=datetime.timezone(datetime.timedelta(0), "+0000"),
            ),
            "docs_job_id": None,
            "repository_id": 85562,
            "created_at": datetime.datetime(
                2022,
                6,
                9,
                21,
                42,
                7,
                72884,
                tzinfo=datetime.timezone(datetime.timedelta(0), "+0000"),
            ),
            "skipped_setup": False,
            "dbt_project_subdirectory": None,
            "group_permissions": [],
            "name": "Analytics",
        },
        {
            "connection": {
                "details": {
                    "project_id": "dbt-tutorial-347100",
                    "timeout_seconds": 300,
                    "private_key_id": "39f4cf7ffd957f8eaaf41e1009401f9c3fb039d6",
                    "client_email": "dbt-user@dbt-tutorial-347100.iam.gserviceaccount.com",
                    "client_id": "106702535903205632050",
                    "auth_uri": "https://accounts.google.com/o/oauth2/auth",
                    "token_uri": "https://oauth2.googleapis.com/token",
                    "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
                    "client_x509_cert_url": "https://www.googleapis.com/robot/v1/metadata/x509/dbt-user%40dbt-tutorial-347100.iam.gserviceaccount.com",
                    "priority": "interactive",
                    "retries": 1,
                    "scopes": [
                        "https://www.googleapis.com/auth/bigquery",
                        "https://www.googleapis.com/auth/cloud-platform",
                        "https://www.googleapis.com/auth/drive",
                    ],
                    "location": None,
                    "maximum_bytes_billed": None,
                    "execution_project": None,
                    "impersonate_service_account": None,
                    "job_retry_deadline_seconds": None,
                    "job_creation_timeout_seconds": None,
                    "is_configured_for_oauth": False,
                },
                "type": "bigquery",
                "id": 79281,
                "account_id": 72449,
                "state": 1,
                "updated_at": datetime.datetime(
                    2022,
                    7,
                    25,
                    21,
                    57,
                    25,
                    938494,
                    tzinfo=datetime.timezone(datetime.timedelta(0), "+0000"),
                ),
                "created_by_id": 89392,
                "project_id": 134905,
                "created_by_service_token_id": None,
                "created_at": datetime.datetime(
                    2022,
                    7,
                    25,
                    21,
                    57,
                    25,
                    938478,
                    tzinfo=datetime.timezone(datetime.timedelta(0), "+0000"),
                ),
                "name": "Bigquery",
            },
            "docs_job": None,
            "freshness_job_id": None,
            "freshness_job": None,
            "id": 134905,
            "connection_id": 79281,
            "account_id": 72449,
            "repository": {
                "gitlab": None,
                "deploy_key_id": 84339,
                "account_id": 72449,
                "repository_credentials_id": None,
                "git_provider_id": 33120,
                "git_clone_strategy": "github_app",
                "deploy_key": {
                    "id": 84339,
                    "account_id": 72449,
                    "state": 1,
                    "public_key": "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAACAQCQ6NWTsb48rEVFc7/Ihsz1XQ6dFa0TQEntuDu6uOPUxBVadup83vhKNqI9xRuOnDDEm/YqhcUZqx6OFB0BFmxSeEoRwTkDkURfBKh1nQ1kSbq0xbKwoRXyvwyfqs1HeTEtSZuPcKelMOWK3R5evCE3ZcKvLQw6oHf32JhHThuaOw/AG5OtvSLlWSxlQ6b7hW64K2kdsf+KODnD9+6PhEMludcBtIwTcZQdhHbyz/UMo3fGas59+LsxhBzjVyF1jmi3XrKLHJ+5oTlvz+Jr2EI2Srvftcciy/DvyGvG/biqrG1KjrCYoqzwjCFhBy4jJBU421QTjBtYLABVJC9hEtlsSFCFMJMubUhKELo3HhX6vbn5E8TF+UCP7wc7JEACyICzxeROk2QAb0aV7gmL/Zrc6mVNou5Cb5mVuv5NIUBjpx7m69rsHRR+lb4H41gZCeeh7dQmV08rZk0L47n8Clcuc55v99o8H7ekk+I9dOhHpeKvbW90XMGtDnZZLfw+o2wG/9S4js1b142nvFP55TfAb6+wOa4W+ZidaH5E3Mp41uk2wgLE2CsRVjAP5vM3OFTU6iyDY/obOMERNEhThKUS83ixR67s5haoWLRy+q36SvLn1oYQYlJ0UZp3SxZUnUTV1m20L2oQ4jAhnr5j3zzsNWCZLW4SxgdOq7Mi16Bqow==",
                },
                "full_name": "preset-io/dbt-integration-blog-post",
                "updated_at": datetime.datetime(
                    2022,
                    7,
                    25,
                    21,
                    57,
                    36,
                    748299,
                    tzinfo=datetime.timezone(datetime.timedelta(0), "+0000"),
                ),
                "github_repo": "preset-io/dbt-integration-blog-post",
                "pull_request_url_template": "https://github.com/preset-io/dbt-integration-blog-post/compare/{{destination}}...{{source}}",
                "created_at": datetime.datetime(
                    2022,
                    7,
                    25,
                    21,
                    57,
                    36,
                    748283,
                    tzinfo=datetime.timezone(datetime.timedelta(0), "+0000"),
                ),
                "git_provider": None,
                "id": 85563,
                "remote_url": "git://github.com/preset-io/dbt-integration-blog-post.git",
                "state": 1,
                "github_installation_id": 27693839,
                "project_id": 134905,
                "remote_backend": "github",
                "name": "dbt-integration-blog-post",
            },
            "state": 1,
            "updated_at": datetime.datetime(
                2022,
                8,
                1,
                23,
                49,
                3,
                645162,
                tzinfo=datetime.timezone(datetime.timedelta(0), "+0000"),
            ),
            "docs_job_id": None,
            "repository_id": 85563,
            "created_at": datetime.datetime(
                2022,
                7,
                25,
                21,
                56,
                3,
                146661,
                tzinfo=datetime.timezone(datetime.timedelta(0), "+0000"),
            ),
            "skipped_setup": False,
            "dbt_project_subdirectory": "jaffle_shop",
            "group_permissions": [],
            "name": "jaffle_shop",
        },
    ]


def test_dbt_client_get_jobs(requests_mock: Mocker) -> None:
    """
    Test the ``get_jobs`` method.
    """
    requests_mock.get(
        "https://cloud.getdbt.com/api/v2/accounts/72449/jobs/",
        json={
            "status": {
                "code": 200,
                "is_success": True,
                "user_message": "Success!",
                "developer_message": "",
            },
            "data": [
                {
                    "execution": {"timeout_seconds": 0},
                    "generate_docs": True,
                    "run_generate_sources": False,
                    "id": 108380,
                    "account_id": 72449,
                    "project_id": 134905,
                    "environment_id": 107605,
                    "name": "Test job",
                    "dbt_version": "1.0.0",
                    "created_at": "2022-07-25T22:00:11.943460+00:00",
                    "updated_at": "2022-07-26T22:36:23.862370+00:00",
                    "execute_steps": ["dbt run", "dbt test"],
                    "state": 1,
                    "deactivated": False,
                    "run_failure_count": 0,
                    "deferring_job_definition_id": None,
                    "lifecycle_webhooks": False,
                    "lifecycle_webhooks_url": None,
                    "triggers": {
                        "github_webhook": False,
                        "git_provider_webhook": False,
                        "custom_branch_only": False,
                        "schedule": True,
                    },
                    "settings": {"threads": 4, "target_name": "default"},
                    "schedule": {
                        "cron": "0 * * * *",
                        "date": {"type": "every_day"},
                        "time": {"type": "every_hour", "interval": 1},
                    },
                    "is_deferrable": False,
                    "generate_sources": False,
                    "cron_humanized": "Every hour",
                    "next_run": "2022-07-26T23:00:00+00:00",
                    "next_run_humanized": "2 weeks, 2 days",
                },
            ],
            "extra": {
                "filters": {"limit": 100, "offset": 0, "account_id": 72449},
                "order_by": "id",
                "pagination": {"count": 1, "total_count": 1},
            },
        },
    )
    auth = Auth()
    client = DBTClient(auth)
    assert client.get_jobs(72449) == [
        {
            "id": 108380,
            "deactivated": False,
            "triggers": {
                "custom_branch_only": False,
                "git_provider_webhook": False,
                "github_webhook": False,
                "schedule": True,
            },
            "next_run_humanized": "2 weeks, 2 days",
            "generate_sources": False,
            "execution": {"timeout_seconds": 0},
            "environment_id": 107605,
            "created_at": datetime.datetime(
                2022,
                7,
                25,
                22,
                0,
                11,
                943460,
                tzinfo=datetime.timezone(datetime.timedelta(0), "+0000"),
            ),
            "account_id": 72449,
            "state": 1,
            "deferring_job_definition_id": None,
            "generate_docs": True,
            "cron_humanized": "Every hour",
            "next_run": datetime.datetime(
                2022,
                7,
                26,
                23,
                0,
                tzinfo=datetime.timezone(datetime.timedelta(0), "+0000"),
            ),
            "run_failure_count": 0,
            "lifecycle_webhooks": False,
            "settings": {"threads": 4, "target_name": "default"},
            "execute_steps": ["dbt run", "dbt test"],
            "project_id": 134905,
            "is_deferrable": False,
            "schedule": {
                "cron": "0 * * * *",
                "time": {"interval": 1, "type": "every_hour"},
                "date": {"type": "every_day"},
            },
            "run_generate_sources": False,
            "dbt_version": "1.0.0",
            "updated_at": datetime.datetime(
                2022,
                7,
                26,
                22,
                36,
                23,
                862370,
                tzinfo=datetime.timezone(datetime.timedelta(0), "+0000"),
            ),
            "lifecycle_webhooks_url": None,
            "name": "Test job",
        },
    ]


def test_dbt_client_get_models(mocker: MockerFixture) -> None:
    """
    Test the ``get_models`` method.
    """
    GraphqlClient = mocker.patch("preset_cli.api.clients.dbt.GraphqlClient")
    GraphqlClient().execute.return_value = {
        "data": {
            "models": [
                {
                    "uniqueId": "model.jaffle_shop.customers",
                    "name": "customers",
                    "database": "dbt-tutorial-347100",
                    "schema": "dbt_beto",
                    "description": "One record per customer",
                    "meta": {"superset": {"cache_timeout": 600}},
                },
                {
                    "uniqueId": "model.jaffle_shop.stg_customers",
                    "name": "stg_customers",
                    "database": "dbt-tutorial-347100",
                    "schema": "dbt_beto",
                    "description": "This model cleans up customer data",
                    "meta": {},
                },
                {
                    "uniqueId": "model.jaffle_shop.stg_orders",
                    "name": "stg_orders",
                    "database": "dbt-tutorial-347100",
                    "schema": "dbt_beto",
                    "description": "This model cleans up order data",
                    "meta": {},
                },
                {
                    "uniqueId": "model.metrics.dbt_metrics_default_calendar",
                    "name": "dbt_metrics_default_calendar",
                    "database": "dbt-tutorial-347100",
                    "schema": "dbt_beto",
                    "description": "",
                    "meta": {},
                },
            ],
        },
    }
    auth = Auth()
    client = DBTClient(auth)
    assert client.get_models(108380) == [
        {
            "schema": "dbt_beto",
            "unique_id": "model.jaffle_shop.customers",
            "description": "One record per customer",
            "meta": {"superset": {"cache_timeout": 600}},
            "database": "dbt-tutorial-347100",
            "name": "customers",
        },
        {
            "schema": "dbt_beto",
            "unique_id": "model.jaffle_shop.stg_customers",
            "description": "This model cleans up customer data",
            "meta": {},
            "database": "dbt-tutorial-347100",
            "name": "stg_customers",
        },
        {
            "schema": "dbt_beto",
            "unique_id": "model.jaffle_shop.stg_orders",
            "description": "This model cleans up order data",
            "meta": {},
            "database": "dbt-tutorial-347100",
            "name": "stg_orders",
        },
        {
            "schema": "dbt_beto",
            "unique_id": "model.metrics.dbt_metrics_default_calendar",
            "description": "",
            "meta": {},
            "database": "dbt-tutorial-347100",
            "name": "dbt_metrics_default_calendar",
        },
    ]


def test_dbt_client_get_metrics(mocker: MockerFixture) -> None:
    """
    Test the ``get_metrics`` method.
    """
    GraphqlClient = mocker.patch("preset_cli.api.clients.dbt.GraphqlClient")
    GraphqlClient().execute.return_value = {
        "data": {
            "metrics": [
                {
                    "uniqueId": "metric.jaffle_shop.new_customers",
                    "name": "new_customers",
                    "label": "New Customers",
                    "type": "count",
                    "sql": "customer_id",
                    "filters": [
                        {"field": "number_of_orders", "operator": ">", "value": "0"},
                    ],
                    "dependsOn": ["model.jaffle_shop.customers"],
                    "description": "The number of paid customers using the product",
                    "meta": {},
                },
            ],
        },
    }
    auth = Auth()
    client = DBTClient(auth)
    assert client.get_metrics(108380) == [
        {
            "meta": {},
            "name": "new_customers",
            "type": "count",
            "label": "New Customers",
            "unique_id": "metric.jaffle_shop.new_customers",
            "description": "The number of paid customers using the product",
            "sql": "customer_id",
            "depends_on": ["model.jaffle_shop.customers"],
            "filters": [{"operator": ">", "value": "0", "field": "number_of_orders"}],
        },
    ]


def test_dbt_client_get_database_name(mocker: MockerFixture) -> None:
    """
    Test the ``get_database_name`` method.
    """
    GraphqlClient = mocker.patch("preset_cli.api.clients.dbt.GraphqlClient")
    GraphqlClient().execute.return_value = {
        "data": {
            "models": [
                {
                    "uniqueId": "model.jaffle_shop.customers",
                    "name": "customers",
                    "database": "dbt-tutorial-347100",
                    "schema": "dbt_beto",
                    "description": "One record per customer",
                    "meta": {"superset": {"cache_timeout": 600}},
                },
                {
                    "uniqueId": "model.jaffle_shop.stg_customers",
                    "name": "stg_customers",
                    "database": "dbt-tutorial-347100",
                    "schema": "dbt_beto",
                    "description": "This model cleans up customer data",
                    "meta": {},
                },
                {
                    "uniqueId": "model.jaffle_shop.stg_orders",
                    "name": "stg_orders",
                    "database": "dbt-tutorial-347100",
                    "schema": "dbt_beto",
                    "description": "This model cleans up order data",
                    "meta": {},
                },
                {
                    "uniqueId": "model.metrics.dbt_metrics_default_calendar",
                    "name": "dbt_metrics_default_calendar",
                    "database": "dbt-tutorial-347100",
                    "schema": "dbt_beto",
                    "description": "",
                    "meta": {},
                },
            ],
        },
    }
    auth = Auth()
    client = DBTClient(auth)
    assert client.get_database_name(108380) == "dbt-tutorial-347100"


def test_dbt_client_get_database_name_no_models(mocker: MockerFixture) -> None:
    """
    Test the ``get_database_name`` method when there are no models.
    """
    GraphqlClient = mocker.patch("preset_cli.api.clients.dbt.GraphqlClient")
    GraphqlClient().execute.return_value = {"data": {"models": []}}
    auth = Auth()
    client = DBTClient(auth)

    with pytest.raises(Exception) as excinfo:
        client.get_database_name(108380)
    assert str(excinfo.value) == "No models found, can't determine database name"


def test_dbt_client_get_database_name_multiple(mocker: MockerFixture) -> None:
    """
    Test the ``get_database_name`` method when there are multiple databases.

    This shouldn't happen.
    """
    GraphqlClient = mocker.patch("preset_cli.api.clients.dbt.GraphqlClient")
    GraphqlClient().execute.return_value = {
        "data": {
            "models": [
                {
                    "uniqueId": "model.jaffle_shop.customers",
                    "name": "customers",
                    "database": "database_two",
                    "schema": "dbt_beto",
                    "description": "One record per customer",
                    "meta": {"superset": {"cache_timeout": 600}},
                },
                {
                    "uniqueId": "model.jaffle_shop.stg_customers",
                    "name": "stg_customers",
                    "database": "database_one",
                    "schema": "dbt_beto",
                    "description": "This model cleans up customer data",
                    "meta": {},
                },
            ],
        },
    }
    auth = Auth()
    client = DBTClient(auth)

    with pytest.raises(Exception) as excinfo:
        client.get_database_name(108380)
    assert str(excinfo.value) == "Multiple databases found"



"""
Custom exceptions.
"""

from enum import Enum
from typing import Any, Dict, List

from typing_extensions import TypedDict


class ErrorLevel(str, Enum):
    """
    Levels of errors that can exist within Superset.
    """

    INFO = "info"
    WARNING = "warning"
    ERROR = "error"


class ErrorPayload(TypedDict, total=False):
    """
    A SIP-40 error payload.
    """

    message: str
    error_type: str  # import SupersetErrorType from Superset?
    level: ErrorLevel
    extra: Dict[str, Any]


class SupersetError(Exception):
    """
    A SIP-40 compliant exception.
    """

    def __init__(self, errors: List[ErrorPayload]):
        super().__init__()
        self.errors = errors


class DatabaseNotFoundError(SupersetError):
    """
    Exception when no database is found.
    """

    def __init__(self):
        super().__init__(
            [
                {
                    "message": "Database not found",
                    "error_type": "DATABASE_NOT_FOUND_ERROR",
                    "level": ErrorLevel.ERROR,
                },
            ],
        )



"""
Test for ``preset_cli.cli.superset.sync.dbt.lib``.
"""
# pylint: disable=invalid-name

import json
import math
from pathlib import Path
from typing import List

import pytest
from pyfakefs.fake_filesystem import FakeFilesystem
from pytest_mock import MockerFixture

from preset_cli.api.clients.dbt import ModelSchema
from preset_cli.cli.superset.sync.dbt.lib import (
    apply_select,
    as_number,
    build_sqlalchemy_params,
    env_var,
    filter_models,
    load_profiles,
)


def test_build_sqlalchemy_params_postgres(mocker: MockerFixture) -> None:
    """
    Test ``build_sqlalchemy_params`` for PostgreSQL.
    """
    _logger = mocker.patch("preset_cli.cli.superset.sync.dbt.lib._logger")
    config = {
        "type": "postgres",
        "user": "username",
        "password": "password123",
        "host": "localhost",
        "port": 5432,
        "dbname": "db",
    }
    assert build_sqlalchemy_params(config) == {
        "sqlalchemy_uri": "postgresql+psycopg2://username:password123@localhost:5432/db",
    }
    _logger.warning.assert_not_called()
    config["search_path"] = "test_schema"
    build_sqlalchemy_params(config)
    _logger.warning.assert_called_with(
        "Specifying a search path is not supported in Apache Superset",
    )


def test_build_sqlalchemy_params_bigquery(fs: FakeFilesystem) -> None:
    """
    Test ``build_sqlalchemy_params`` for BigQuery.
    """
    fs.create_file(
        "/path/to/credentials.json",
        contents=json.dumps({"Hello": "World!"}),
    )
    config = {
        "type": "bigquery",
        "project": "my_project",
        "keyfile": "/path/to/credentials.json",
    }
    assert build_sqlalchemy_params(config) == {
        "sqlalchemy_uri": "bigquery://my_project/",
        "encrypted_extra": json.dumps({"credentials_info": {"Hello": "World!"}}),
    }


def test_build_sqlalchemy_params_bigquery_with_priority(fs: FakeFilesystem) -> None:
    """
    Test ``build_sqlalchemy_params`` for BigQuery with priority parameter.

    Parameter should be uppercase.
    """
    fs.create_file(
        "/path/to/credentials.json",
        contents=json.dumps({"Hello": "World!"}),
    )
    config = {
        "type": "bigquery",
        "project": "my_project",
        "keyfile": "/path/to/credentials.json",
        "priority": "interactive",
    }
    assert build_sqlalchemy_params(config) == {
        "sqlalchemy_uri": "bigquery://my_project/?priority=INTERACTIVE",
        "encrypted_extra": json.dumps({"credentials_info": {"Hello": "World!"}}),
    }


def test_build_sqlalchemy_params_bigquery_no_keyfile() -> None:
    """
    Test ``build_sqlalchemy_params`` for BigQuery with priority parameter.

    Parameter should be uppercase.
    """
    config = {
        "type": "bigquery",
        "project": "my_project",
    }
    with pytest.raises(Exception) as excinfo:
        build_sqlalchemy_params(config)
    assert (
        str(excinfo.value)
        == "Only service account auth is supported, you MUST pass `keyfile`."
    )


def test_build_snowflake_sqlalchemy_params() -> None:
    """
    Test ``build_snowflake_sqlalchemy_params`` for Snowflake.
    """
    config = {
        "type": "snowflake",
        "account": "abc123.eu-west-1.aws",
        "user": "jdoe",
        "password": "secret",
        "role": "admin",
        "database": "default",
        "warehouse": "dunder-mifflin",
    }
    assert build_sqlalchemy_params(config) == {
        "sqlalchemy_uri": (
            "snowflake://jdoe:secret@abc123.eu-west-1.aws/default?"
            "role=admin&warehouse=dunder-mifflin"
        ),
    }


def test_build_sqlalchemy_params_unsupported() -> None:
    """
    Test ``build_sqlalchemy_params`` for databases currently unsupported.
    """
    config = {"type": "mysql"}
    with pytest.raises(Exception) as excinfo:
        build_sqlalchemy_params(config)
    assert str(excinfo.value) == (
        "Unable to build a SQLAlchemy URI for a target of type mysql. Please file "
        "an issue at https://github.com/preset-io/backend-sdk/issues/new?"
        "labels=enhancement&title=Backend+for+mysql."
    )


def test_env_var(monkeypatch: pytest.MonkeyPatch) -> None:
    """
    Test the ``env_var`` implementation.
    """
    monkeypatch.setenv("MY_USER", "Nanna")

    assert env_var("MY_USER") == "Nanna"
    assert env_var("YOUR_USER", "Jane Doe") == "Jane Doe"
    with pytest.raises(Exception) as excinfo:
        env_var("YOUR_USER")
    assert str(excinfo.value) == "Env var required but not provided: 'YOUR_USER'"


def test_as_number() -> None:
    """
    Test ``as_number`` macro.
    """
    assert as_number("1.0") == 1
    assert as_number("1.1") == 1.1
    assert as_number("2") == 2
    assert math.isnan(as_number("nan"))
    with pytest.raises(ValueError) as excinfo:
        as_number("invalid")
    assert str(excinfo.value) == "could not convert string to float: 'invalid'"


def test_load_profiles(monkeypatch: pytest.MonkeyPatch, fs: FakeFilesystem) -> None:
    """
    Test ``load_profiles``.
    """
    monkeypatch.setenv("REDSHIFT_HOST", "127.0.0.1")
    monkeypatch.setenv("REDSHIFT_PORT", "1234")
    monkeypatch.setenv("REDSHIFT_USER", "username")
    monkeypatch.setenv("REDSHIFT_PASSWORD", "password123")
    monkeypatch.setenv("REDSHIFT_DATABASE", "db")
    monkeypatch.setenv("THREADS", "3")

    path = Path("/path/to/profiles.yml")
    fs.create_file(
        path,
        contents="""
jaffle_shop:
  outputs:
    dev:
      host: "{{ env_var('REDSHIFT_HOST') | as_text }}"
      port: "{{ env_var('REDSHIFT_PORT') | as_number }}"
      user: "{{ env_var('REDSHIFT_USER') }}"
      pass: "{{ env_var('REDSHIFT_PASSWORD') }}"
      dbname: "{{ env_var('REDSHIFT_DATABASE') }}"
      schema: public
      threads: "{{ env_var('THREADS') | as_native }}"
      type: postgres
      enabled: "{{ (target.name == 'prod') | as_bool }}"
      a_list: [1, 2, 3]
      a_value: 10
  target: dev
    """,
    )

    assert load_profiles(path, "jaffle_shop", "dev") == {
        "jaffle_shop": {
            "outputs": {
                "dev": {
                    "host": "127.0.0.1",
                    "port": 1234,
                    "user": "username",
                    "pass": "password123",
                    "dbname": "db",
                    "schema": "public",
                    "threads": 3,
                    "type": "postgres",
                    "enabled": False,
                    "a_list": [1, 2, 3],
                    "a_value": 10,
                },
            },
            "target": "dev",
        },
    }


def test_filter_models() -> None:
    """
    Test ``filter_models``.
    """
    one = {
        "name": "one",
        "tags": ["test"],
        "unique_id": "model.one",
        "depends_on": ["source.zero"],
        "children": ["model.two"],
    }
    two = {
        "name": "two",
        "tags": [],
        "unique_id": "model.two",
        "depends_on": ["model.one", "model.three"],
        "children": [],
    }
    three = {
        "name": "three",
        "tags": [],
        "unique_id": "model.three",
        "depends_on": ["source.zero"],
        "children": ["model.two"],
    }
    models: List[ModelSchema] = [one, two, three]  # type: ignore

    assert {model["name"] for model in filter_models(models, "one")} == {"one"}
    assert {model["name"] for model in filter_models(models, "one+")} == {
        "one",
        "two",
    }
    assert {model["name"] for model in filter_models(models, "+two")} == {
        "one",
        "two",
        "three",
    }
    assert {model["name"] for model in filter_models(models, "tag:test")} == {"one"}
    assert {model["name"] for model in filter_models(models, "@one")} == {
        "one",
        "two",
        "three",
    }

    with pytest.raises(NotImplementedError) as excinfo:
        filter_models(models, "invalid")
    assert str(excinfo.value) == (
        "Unable to parse the selection invalid. Please file an issue at "
        "https://github.com/preset-io/backend-sdk/issues/new?"
        "labels=enhancement&title=dbt+select+invalid."
    )


def test_filter_models_seen() -> None:
    """
    Test that ``filter_models`` dedupes models.
    """
    one = {
        "name": "one",
        "tags": [],
        "unique_id": "model.one",
        "depends_on": ["source.zero"],
        "children": ["model.two", "model.three"],
    }
    two = {
        "name": "two",
        "tags": [],
        "unique_id": "model.two",
        "depends_on": ["model.one"],
        "children": ["model.four"],
    }
    three = {
        "name": "three",
        "tags": [],
        "unique_id": "model.three",
        "depends_on": ["model.one"],
        "children": ["model.four"],
    }
    four = {
        "name": "four",
        "tags": [],
        "unique_id": "model.four",
        "depends_on": ["model.two", "model.three"],
        "children": [],
    }
    models: List[ModelSchema] = [one, two, three, four]  # type: ignore

    assert {model["name"] for model in filter_models(models, "+four")} == {
        "one",
        "two",
        "three",
        "four",
    }
    assert {model["name"] for model in filter_models(models, "one+")} == {
        "one",
        "two",
        "three",
        "four",
    }
    assert {model["name"] for model in filter_models(models, "1+four")} == {
        "two",
        "three",
        "four",
    }
    assert {model["name"] for model in filter_models(models, "one+1")} == {
        "one",
        "two",
        "three",
    }


def test_apply_select() -> None:
    """
    Test ``apply_select``.
    """
    one = {
        "name": "one",
        "tags": ["test"],
        "unique_id": "model.one",
        "depends_on": ["source.zero"],
        "children": ["model.two"],
    }
    two = {
        "name": "two",
        "tags": [],
        "unique_id": "model.two",
        "depends_on": ["model.one", "model.three"],
        "children": [],
    }
    three = {
        "name": "three",
        "tags": [],
        "unique_id": "model.three",
        "depends_on": ["source.zero"],
        "children": ["model.two"],
    }
    models: List[ModelSchema] = [one, two, three]  # type: ignore

    assert {model["name"] for model in apply_select(models, ("one", "two"), ())} == {
        "one",
        "two",
    }
    assert {model["name"] for model in apply_select(models, ("+two+",), ())} == {
        "one",
        "two",
        "three",
    }
    assert {
        model["name"] for model in apply_select(models, ("+two+,tag:test",), ())
    } == {
        "one",
    }
    assert {
        model["name"] for model in apply_select(models, ("tag:test,+two+",), ())
    } == {
        "one",
    }

    # test exclude
    assert {
        model["name"]
        for model in apply_select(models, ("+two+",), ("three", "tag:test"))
    } == {
        "two",
    }
